"""This module defines the TopoEntity class and its subclasses Bond, Angle, Dihedral, and Improper."""
from collections import namedtuple
import numpy as np
from Bio.PDB.Selection import unfold_entities

class TopoEntity:
    """A TopoEntity is a base class for Topology entities such as Bonds, Angles, 
    Dihedrals, and Impropers."""
    RED = '\033[91m'
    ENDC = '\033[0m'

    def _create_full_id(self):
        """Create unique id by comparing the end atoms to decide if the sequence need to be flipped"""
        # we only need the last 3 elements of the full id (chain, resi, name)
        atom_ids = tuple(a.get_full_id()[-3:] for a in self)
        if atom_ids[0] > atom_ids[-1]:
            atom_ids = tuple(reversed(atom_ids))
        return atom_ids

    def __getnewargs__(self):
        "Support for pickle protocol 2: http://docs.python.org/2/library/pickle.html#pickling-and-unpickling-normal-class-instances"
        return *self, self.param

    def __getstate__(self):
        """
        Additional support for pickle since parent class implements its own __getstate__
        so pickle does not store or restore the type and order, python 2 problem only
        https://www.python.org/dev/peps/pep-0307/#case-3-pickling-new-style-class-instances-using-protocol-2
        """
        return self.__dict__

    def __hash__(self) -> int:
        return hash(self.full_id)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.full_id == other.full_id
 
    def __deepcopy__(self, memo):
        return type(self)(*self, **self.__dict__)

    def get_atom_types(self):
        """Return the atom types of the atoms in this entity"""
        return tuple(a.topo_definition.atom_type for a in self)

    def color_missing(self, atom):
        """Return a string representation of an atom with missing coordinates colored red"""
        if atom.coord is None:
            return f"{self.RED}{atom.name:>4s}{self.ENDC}"
        else:
            return f"{atom.name:>4s}"

    @property
    def parent(self):
        """Return the parent Residue objects in a list"""
        return tuple(unfold_entities(list(self), 'R'))

BondTuple = namedtuple('Bond', ['atom1', 'atom2'])
class Bond(TopoEntity, BondTuple):
    """A Bond object represents a bond between two Atoms within a Topology.

    This class extends tuple, and may be interpreted as a 2-element tuple of Atom objects.
    It also has fields that can optionally be used to describe the bond order and type of bond."""

    bond_order_dict = {'single':1, 'double':2, 'triple':3, 'aromatic':2}
    def __new__(cls, atom1, atom2, bond_type=None, param=None):
        """Create a new Bond. """
        bond = super(Bond, cls).__new__(cls, atom1, atom2)
        bond.type = bond_type
        bond.order = cls.bond_order_dict.get(type)
        bond.param = param
        bond.full_id = bond._create_full_id()
        bond.atom_types = bond.get_atom_types()
        return bond

    def __getnewargs__(self):
        "Support for pickle protocol 2: http://docs.python.org/2/library/pickle.html#pickling-and-unpickling-normal-class-instances"
        return *self, self.type, self.param
    
    def __repr__(self):
        a, b = (self.color_missing(atom) for atom in self)
        s = f"<Bond({a}, {b})"
        if self.type is not None:
            s += f" type={self.type}"
        if self.order is not None:
            s += f" order={self.order:d}"
        if self.length is not None:
            s += f" length={self.length:.2f}"
        s += ">"
        return s

    def __deepcopy__(self, memo):
        return Bond(*self, self.type, self.param)
    
    @property
    def length(self):
        """return the current bond length"""
        if self[0].coord is None or self[1].coord is None:
            return None
        return (((self[0].coord - self[1].coord)**2).sum())**0.5
    
    @property
    def kb(self):
        """return the bond force constant"""
        if self.param is None:
            return None
        return self.param.kb
    
    @property
    def b0(self):
        """return the bond equilibrium length"""
        if self.param is None:
            return None
        return self.param.b0

AngleTuple = namedtuple('Angle', ['atom1', 'atom2', 'atom3'])
class Angle(TopoEntity, AngleTuple):
    """An Angle object represents an angle between three Atoms within a Topology.

    This class extends tuple, and may be interpreted as a 2-element tuple of Bond objects.
    """
    
    def __new__(cls, atom1, atom2, atom3, param=None):
        """Create a new Entity. """
        angle = super(Angle, cls).__new__(cls, atom1, atom2, atom3)
        angle.param = param
        angle.full_id = angle._create_full_id()
        angle.atom_types = angle.get_atom_types()
        return angle
    
    def __repr__(self):
        a, b, c = (self.color_missing(atom) for atom in self)
        s = f"<Angle({a:>4s}, {b:>4s}, {c:>4s})"
        if self.angle is not None:
            s += f" angle={self.angle:.2f}"
        s += ">"
        return s
    
    @property
    def angle(self):
        """return the current angle in degrees"""
        a, b, c = (atom.coord for atom in self)
        for coord in (a, b, c):
            if coord is None:
                return None
        ba = a - b
        bc = c - b
        theta = np.arccos(np.dot(ba, bc) / (np.linalg.norm(ba) * np.linalg.norm(bc)))
        return np.rad2deg(theta)
    
    @property
    def ktheta(self):
        """return the angle force constant"""
        if self.param is None:
            return None
        return self.param.ktheta
    
    @property
    def theta0(self):
        """return the equilibrium angle"""
        if self.param is None:
            return None
        return self.param.theta0
    
DiheTuple = namedtuple('Dihedral', ['i', 'j', 'k', 'l'])
class Dihedral(TopoEntity, DiheTuple):
    """A Dihedral object represents a dihedral angle between four Atoms within a Topology.

    This class extends tuple, and may be interpreted as a 4-element tuple (i, j, k, l) of 
    Atom objects.
    """

    def __new__(cls, atom_i, atom_j, atom_k, atom_l, param=None):
        """Create a new Entity. """
        dihedral = super(Dihedral, cls).__new__(cls, atom_i, atom_j, atom_k, atom_l)
        dihedral.param = param
        dihedral.full_id = dihedral._create_full_id()
        dihedral.atom_types = dihedral.get_atom_types()
        return dihedral

    def __repr__(self):
        a, b, c, d = (self.color_missing(atom) for atom in self)
        s = f"<Dihedral({a:>4s}, {b:>4s}, {c:>4s}, {d:>4s})"
        if self.angle is not None:
            s += f" angle={self.angle:.2f}"
        s += ">"
        return s

    ## TODO: implement dihedral angle calculation
    @property
    def angle(self):
        """return the current dihedral in degrees"""
        a, b, c, d = (atom.coord for atom in self)
        for coord in (a, b, c, d):
            if coord is None:
                return None
        # ba = a - b
        # dc = d - c
        
        return 0.000


ImprTuple = namedtuple('Improper', ['i', 'j', 'k', 'l'])
class Improper(TopoEntity, ImprTuple):
    """A Dihedral object represents a dihedral angle between four Atoms within a Topology.

    This class extends tuple, and may be interpreted as a 4-element tuple (i, j, k, l) of 
    Atom objects, where i is the center atom.
    """

    def __new__(cls, atom_i, atom_j, atom_k, atom_l, param=None):
        """Create a new Entity. """
        improper = super(Improper, cls).__new__(cls, atom_i, atom_j, atom_k, atom_l)
        improper.param = param
        improper.full_id = improper._create_full_id()
        improper.atom_types = improper.get_atom_types()
        return improper
    
    def __repr__(self):
        a, b, c, d = (self.color_missing(atom) for atom in self)
        s = f"<Improper({a:>4s}, {b:>4s}, {c:>4s}, {d:>4s})"
        if self.angle is not None:
            s += f" angle={self.angle:.2f})"
        s += ">"
        return s
    
    @property
    def angle(self):
        """return the current improper in degrees"""
        a, b, c, d = (atom.coord for atom in self)
        for coord in (a, b, c, d):
            if coord is None:
                return None
        return 0.000

CmapTuple = namedtuple('CMap', ['dihe1', 'dihe2'])
class CMap(TopoEntity, CmapTuple):

    def __new__(cls, dihe1, dihe2, param=None):
        """Create a new Entity. """
        cmap = super(CMap, cls).__new__(cls, dihe1, dihe2)
        cmap.param = param
        cmap.full_id = sorted((dihe1.full_id, dihe2.full_id))
        return cmap
    
    def __repr__(self):
        dihe1, dihe2 = self
        s = f"<Cross-Term({repr(dihe1)}, {repr(dihe2)})"
        if self.angle is not None:
            s += f" angle={self.angle:.2f})"
        s += ">"
        return s
    
    def get_atom_types(self):
        """Return the atom types of the atoms in this entity"""
        dihe1, dihe2 = self
        atom_types1 = dihe1.get_atom_types()
        atom_types2 = dihe2.get_atom_types()
        return tuple(sorted((atom_types1, atom_types2)))
    
    @property
    def angle(self):
        """return the current improper in degrees"""
        a, b, c, d = (atom.coord for atom in self)
        for coord in (a, b, c, d):
            if coord is None:
                return None
        return 0.000


import warnings
from typing import List, Tuple
from Bio.PDB.Atom import Atom
import numpy as np
from numpy.linalg import norm
from scipy.spatial.transform import Rotation as R
from scipy.spatial.distance import pdist, squareform

class CoordManipulator:
    def __init__(self) -> None:
        self.entity = None
        self.include_alt = None
        self._atoms = None
        self.coords = None
        self.dist_matrix = None
        self.end_i, self.end_j = None, None
        self.op_mat = None
        self._convex_hull = None

    def load_entity(self, entity, include_alt=False) -> None:
        """Load a structure entity to find translation and rotation operations to
        orient the structure such that the major axis is along the x-axis, and the
        farthest point from the major axis is along the y-axis."""
        self.entity = entity
        self.include_alt = include_alt
        self._atoms, self.coords = self._extract_atoms_and_coords(
            self.entity, include_alt=self.include_alt
        )
        self.dist_matrix = squareform(pdist(self.coords))
        # pair of indices of the farthest atoms in the structure
        self.end_i, self.end_j = self._find_farthest_atom_indices()
        self.m_translation, self.m_rotation = None, None

    def _extract_atoms_and_coords(self, entity, include_alt) -> Tuple[List[Atom], np.array]:
        coords = []
        atoms = []
        for atom in entity.get_atoms(include_alt=include_alt):
            coords.append(atom.coord)
            atoms.append(atom)
        return atoms, np.asarray(coords)

    def _find_farthest_atom_indices(self) -> Tuple[int, int]:
        idx_pair = np.unravel_index(
            np.argmax(self.dist_matrix, axis=None),
            self.dist_matrix.shape
        )
        return idx_pair

    def get_transformation_matrix(self) -> Tuple[np.array, np.array]:
        """Return the 4x4 transformation matrix as numpy arrays."""
        if self.op_mat is None:
            self.op_mat = self._find_transformation_operators()
        return self.op_mat

    def _find_transformation_operators(self) -> None:
        a1, a2 = self.end_i, self.end_j
        # coordinates of the farthest atom pair
        o1, o2 = self.coords[a1], self.coords[a2]
        # Translation operator move the center of structure to the origin
        # the center is defined by the midpoint of the major axis 
        # (the line between farthest atoms)
        translation = -(o1 + o2)/2
        major_axis = o1 - o2
        # Translate the structure to the origin
        temp_coords = self.coords + translation
        # coordinates of the farthest atom pair after translation
        c1, c2 = temp_coords[a1], temp_coords[a2]
        # Find the third atom that is the farthest from the major axis
        dist_yz = np.abs(
            norm(np.cross(major_axis, temp_coords), axis = 1)/norm(major_axis)
        )
        # coordinates of the third atom after translation but before rotation
        a3 = dist_yz.argmax()
        c3 = temp_coords[a3]

        # coordinates of the third atom after rotation
        c3_y = dist_yz.max()
        # Find the sign of the x-coordinate of the third atom
        sign = np.sign(norm(c3-c1)-norm(c3-c2))
        c3_x = sign*np.sqrt(np.sum(c3**2) - c3_y**2)

        vec_translated = np.vstack([c1, c2, c3])
        vec_rotated = np.array([
            [-norm(c1), 0, 0],
            [norm(c2), 0, 0],
            [c3_x, c3_y, 0]
        ])
        # Estimate the rotation by Kabsch algorithm implemented in scipy
        rot_obj, rssd = R.align_vectors(vec_rotated, vec_translated)
        # Get the rotation operator as a numpy array
        rotation = rot_obj.as_matrix()
        # Test if the rotation operators can be applied to the translated vectors
        assert np.allclose(vec_translated @ rotation.T, vec_rotated)
        # combine the translation and rotation operators
        combined = np.eye(4)
        # since we translate first, the rotation needs to be applied to the translation vector too
        combined[:3, 3] = translation @ rotation.T
        combined[:3, :3] = rotation

        # find the final translation to recenter the structure
        temp_coords = temp_coords @ rotation.T
        recenter = -(temp_coords.max(0) + temp_coords.min(0))/2
        combined[:3, 3] += recenter

        return combined

    @property
    def coord_center(self):
        """Return the center of the coordinates (N, 3). The center is defined as
        the midpoint of the maximum and minimum coordinates of each dimension. 
        Should be (0, 0, 0) after the transformation by `orient_coords`.
        """
        return (self.coords.max(0) + self.coords.min(0))/2

    @property
    def box_dim(self):
        """Return the dimensions of the bounding box of the coordinates (N, 3).
        The three sides of the box are parallel to the x, y, and z axes.
        """
        return self.coords.ptp(0)

    def apply_coords(self, coords) -> np.array:
        """Apply the transformation to the coordinates (N, 3). Specifically, the 
        transformation is applied as `(coords + translation) @ rotation`, where
        the results are rounded to 4 decimal places."""
        operator = self.get_transformation_matrix()
        homo_coords = np.column_stack((coords, np.ones(len(coords))))
        new_coords = (homo_coords @ operator.T)[:, :3]
        # round the coordinates to 4 decimal places
        new_coords = np.around(new_coords, decimals=4, out=None)
        return new_coords

    def apply_entity(self, other_entity) -> None:
        """Apply the same transfermation to another structure entity."""
        atoms, coords = self._extract_atoms_and_coords(
            other_entity, self.include_alt
        )
        new_coords = self.apply_coords(coords)
        for i, atom in enumerate(atoms):
            atom.coord = new_coords[i]

    def _apply_to_loaded_entity(self) -> None:
        self.coords = self.apply_coords(self.coords)
        for i, atom in enumerate(self._atoms):
            atom.coord = self.coords[i]

    def orient_coords(self, apply_to_parent = False) -> None:
        """Apply translation and rotation operations to orient the structure 
        such that the major axis of the structure is placed  along the x-axis, 
        and the farthest point from the major axis is along the y-axis. 
        If apply_to_parent, the same transformation will also be applied to the 
        parent structure entity.
        """
        if self.entity is None:
            raise ValueError('No structure entity loaded!')
        self._find_transformation_operators()
        if not apply_to_parent:
            self._apply_to_loaded_entity()
        elif self.entity.parent is not None:
            self.apply_entity(self.entity.parent)
            # update the coordinates and atoms
            self._atoms, self.coords = self._extract_atoms_and_coords(
                self.entity, self.include_alt
            )
        else:
            self._apply_to_loaded_entity()
            warnings.warn('No parent structure entity exists!')
            
    def orient_coords_octa(self, apply_to_parent=False) -> None:
        """
        Apply an alternative orientation aimed at minimizing the bounding
        truncated octahedron. This implementation uses a PCA-based approach to
        align the principal axes with the coordinate axes.

        Note: If apply_to_parent=True, the transformation is applied to the
        parent entity's atoms. Since atoms are shared objects in the Biopython
        hierarchy (model atoms ARE structure atoms), modifying the loaded
        entity's atoms already updates the parent. The apply_to_parent flag
        is kept for API consistency but the transformation is only applied once.
        """
        if self.entity is None:
            raise ValueError('No structure entity loaded!')
        # Center the coordinates
        center = self.coords.mean(axis=0)
        centered_coords = self.coords - center
        # Compute PCA (using SVD)
        U, S, Vt = np.linalg.svd(centered_coords, full_matrices=False)
        rotation = Vt.T  # principal axes
        new_coords = centered_coords @ rotation
        # Re-center the new coordinates around the origin
        new_center = new_coords.mean(axis=0)
        new_coords -= new_center

        # Build the 4x4 transformation matrix for potential external use
        # The transformation is: (coords - center) @ rotation - new_center
        # Which equals: coords @ rotation - center @ rotation - new_center
        translation = -center @ rotation - new_center
        self.op_mat = np.eye(4)
        self.op_mat[:3, :3] = rotation
        self.op_mat[3, :3] = translation

        self.coords = new_coords
        # Update atoms - since atoms are shared between model and parent structure,
        # this single update affects both. No need to call apply_entity separately.
        for i, atom in enumerate(self._atoms):
            atom.coord = new_coords[i]

    def orient_coords_ortho(self, apply_to_parent=False) -> None:
        """
        Orient coordinates for orthorhombic box - align principal axes and sort by extent.

        Uses PCA to find principal axes, then sorts them so:
        - Largest extent → X axis (box dimension 'a')
        - Medium extent → Y axis (box dimension 'b')
        - Smallest extent → Z axis (box dimension 'c')

        This orientation minimizes box volume for orthorhombic geometry by ensuring
        that box dimensions match the molecule's extent ordering.

        Note: Same parent/child sharing behavior as orient_coords_octa.
        """
        if self.entity is None:
            raise ValueError('No structure entity loaded!')
        
        # Center the coordinates
        center = self.coords.mean(axis=0)
        centered_coords = self.coords - center
        
        # Compute PCA using SVD
        U, S, Vt = np.linalg.svd(centered_coords, full_matrices=False)
        
        # S contains singular values proportional to extents along each principal axis
        # Sort principal components by decreasing extent (largest first → X axis)
        sort_idx = np.argsort(S)[::-1]  # Indices for descending order
        rotation = Vt[sort_idx].T
        
        # Ensure right-handed coordinate system
        if np.linalg.det(rotation) < 0:
            rotation[:, -1] *= -1
        
        # Apply rotation
        new_coords = centered_coords @ rotation
        
        # Re-center the new coordinates around the origin
        new_center = new_coords.mean(axis=0)
        new_coords -= new_center
        
        # Build the 4x4 transformation matrix
        translation = -center @ rotation - new_center
        self.op_mat = np.eye(4)
        self.op_mat[:3, :3] = rotation
        self.op_mat[3, :3] = translation
        
        self.coords = new_coords
        for i, atom in enumerate(self._atoms):
            atom.coord = new_coords[i]

    def orient_coords_hexa(self, apply_to_parent=False) -> None:
        """
        Orient coordinates for hexagonal box - align smallest extent with Z.

        The hexagonal prism has its base in the XY plane (γ=120°).
        For optimal packing, orient molecule so its flattest dimension
        aligns with Z (the prism height), while the two larger dimensions
        span the hexagonal base.

        Axis ordering after orientation:
        - Two largest extents → XY plane (hexagonal base)
        - Smallest extent → Z axis (prism height)

        Note: Same parent/child sharing behavior as orient_coords_octa.
        """
        if self.entity is None:
            raise ValueError('No structure entity loaded!')
        
        # Center the coordinates
        center = self.coords.mean(axis=0)
        centered_coords = self.coords - center
        
        # Compute PCA using SVD
        U, S, Vt = np.linalg.svd(centered_coords, full_matrices=False)
        
        # S contains singular values proportional to extents
        # We want: largest two extents in XY, smallest in Z
        # Sort indices by S value
        sort_idx = np.argsort(S)  # Ascending order (smallest first)
        
        # Reorder: [largest, medium, smallest] → [X, Y, Z]
        # sort_idx[2] is largest, sort_idx[1] is medium, sort_idx[0] is smallest
        reorder = [sort_idx[2], sort_idx[1], sort_idx[0]]
        rotation = Vt[reorder].T
        
        # Ensure right-handed coordinate system
        if np.linalg.det(rotation) < 0:
            rotation[:, -1] *= -1
        
        # Apply rotation
        new_coords = centered_coords @ rotation
        
        # Re-center the new coordinates around the origin
        new_center = new_coords.mean(axis=0)
        new_coords -= new_center
        
        # Build the 4x4 transformation matrix
        translation = -center @ rotation - new_center
        self.op_mat = np.eye(4)
        self.op_mat[:3, :3] = rotation
        self.op_mat[3, :3] = translation
        
        self.coords = new_coords
        for i, atom in enumerate(self._atoms):
            atom.coord = new_coords[i]



from collections import namedtuple
from crimm.IO.RTFParser import comment_parser, skip_line

bond_par = namedtuple('bond_param',['kb', 'b0'])
angle_par = namedtuple('angle_param',['ktheta', 'theta0'])
ub_par = namedtuple('ub_param',['kub', 's0']) # Urey-Bradley
dihe_par = namedtuple('dihedral_param',['kchi', 'n', 'delta'])
impr_par = namedtuple('improper_param',['kpsi', 'psi0'])
cmap_par = namedtuple('cmap_param',['degree', 'values'])
nonbond_par = namedtuple('nonbond_param', ['epsilon', 'rmin_half'])
nonbond14_par = namedtuple('nonbond14_param', ['epsilon', 'rmin_half'])
nbfix_par = namedtuple('nbfix_param', ['emin','rmin'])

def categorize_lines(lines):
    line_dict = {
        'mass': [], 'bonds': [], 'angles': [], 'dihedrals': [],
        'improper': [], 'cmap': [], 'nonbonded': [], 'nbfix':[]
    }
    is_continued = False
    for line in lines:
        l = line.strip().upper()
        if skip_line(l):
            continue
        if l.endswith('-'):
            if is_continued:
                is_continued += l.rstrip('-')
            else:
                is_continued = l.rstrip('-')
            continue
        if is_continued:
            l = is_continued
            is_continued = False
        if l.startswith('ATOMS'):
            cur_section = line_dict['mass']
        elif l.startswith('BONDS'):
            cur_section = line_dict['bonds']
        elif l.startswith('ANGLES'):
            cur_section = line_dict['angles']
        elif l.startswith('DIHEDRALS'):
            cur_section = line_dict['dihedrals']
        elif l.startswith('IMPROPER'):
            cur_section = line_dict['improper']
        elif l.startswith('CMAP'):
            cur_section = line_dict['cmap']
        elif l.startswith('NONBONDED'):
            cur_section = line_dict['nonbonded']
        elif l.startswith('NBFIX'):
            cur_section = line_dict['nbfix']
        elif l.startswith('HBOND'):
            pass
        elif l.startswith('END'):
            break
        else:
            cur_section.append(l)
    return line_dict

def bond_parser(line):
    field_str, desc = comment_parser(line)
    # fields: [atom_type1, atom_type2, Kb: kcal/mole/A**2, b0: A]
    atom1, atom2, kb, b0 = field_str.split()
    return (atom1, atom2), bond_par(float(kb), float(b0))

def angle_parser(line):
    # V(angle) = Ktheta(Theta - Theta0)**2
    # V(Urey-Bradley) = Kub(S - S0)**2
    # fields : [
    #     atom_type1, atom_type2, atom_type3,
    #     Ktheta: kcal/mole/rad**2, Theta0: degrees
    #     Kub: kcal/mole/A**2 (Urey-Bradley), S0: A
    # ]
    field_str, desc = comment_parser(line)
    fields = field_str.split()
    if len(fields) == 7:
        atom1, atom2, atom3, ktheta, theta0, kub, s0 = fields
        return (
            (atom1, atom2, atom3), 
            angle_par(float(ktheta), float(theta0)), 
            ub_par(float(kub), float(s0))
        )
    atom1, atom2, atom3, ktheta, theta0 = fields
    return (
        (atom1, atom2, atom3), 
        angle_par(float(ktheta), float(theta0)), 
        None
    )

def dihedral_parser(line):
    # V(dihedral) = Kchi(1 + cos(n(chi) - delta))
    # fields: [
    #     atom1, atom2, atom3, atom4, 
    #     Kchi: kcal/mole, n: multiplicity, delta: degrees
    # ]
    field_str, desc = comment_parser(line)
    a1, a2, a3, a4, kchi, n, delta = field_str.split()
    return (a1,a2,a3,a4), dihe_par(float(kchi), int(n), float(delta))

def improper_parser(line):
    # V(improper) = Kpsi(psi - psi0)**2
    # fields: [
    #     atom1, atom2, atom3, atom4, 
    #     Kpsi: kcal/mole/rad**2, ignored, psi0: degrees
    # ]
    field_str, desc = comment_parser(line)
    a1, a2, a3, a4, kpsi, _, psi0 = field_str.split()
    return (a1,a2,a3,a4), impr_par(float(kpsi), float(psi0))

def cmap_key_parser(line):
    # fields: [8 atoms, n_cmap_entries]
    fields = line.split()
    return tuple(fields[:8]), int(fields[-1])

def cmap_field_parser(lines, n):
    entries = []
    for line in lines:
        entries.extend([float(entry) for entry in line.split()])
    assert(len(entries) == n**2)
    return entries

def nonbonded_parser(line):
    # V(Lennard-Jones) = Eps,i,j[(Rmin,i,j/ri,j)**12 - 2(Rmin,i,j/ri,j)**6]
    # fields: [
    #     atom, ignored
    #     epsilon: kcal/mole, 
    #     Eps,i,j = sqrt(eps,i * eps,j)
    #     Rmin/2 (Rmin,i,j = Rmin/2,i + Rmin/2,j): A, 
    #     ignored, eps,1-4, Rmin/2,1-4
    # ]
    field_str, desc = comment_parser(line)
    field = field_str.split()
    if len(field) == 7:
        atom, _, eps, rmin_2, _, eps_14, rmin_2_14 = field
        return (
            atom, 
            nonbond_par(float(eps), float(rmin_2)), 
            nonbond14_par(float(eps_14), float(rmin_2_14))
        )
    atom, _, eps, rmin_2 = field
    return atom, nonbond_par(float(eps), float(rmin_2)), None

def nbfix_parser(line):
    # fields:[atom1, atom2, Emin: kcal/mol, Rmin: A]
    field_str, desc = comment_parser(line)
    atom1, atom2, emin, rmin = field_str.split()
    return (atom1, atom2), nbfix_par(float(emin), float(rmin))

def simple_collector(param_line_dict, parser):
    param_dict = {}
    for l in param_line_dict:
        key, val = parser(l)
        if key in param_dict:
            continue
        param_dict[key] = val
    return param_dict

def dihedral_collector(param_line_dict, parser):
    param_dict = {}
    for l in param_line_dict:
        key, val = parser(l)
        if key not in param_dict:
            param_dict[key] = []
        param_dict[key].append(val)
    for value_list in param_dict.values():
        value_list.sort(key=lambda x: x.n)
    return param_dict

def extended_collector(param_line_dict, parser):
    param_dict = {}
    extended_param_dict = {}
    for l in param_line_dict:
        key, val1, val2 = parser(l)
        param_dict[key] = val1
        if val2 is not None:
            extended_param_dict[key] = val2
    return param_dict, extended_param_dict

def parse_line_dict(line_dict):

    bond_dict = simple_collector(line_dict['bonds'], bond_parser)
    angle_dict, urey_bradley_dict = extended_collector(
        line_dict['angles'], angle_parser
    )
    dihe_dict = dihedral_collector(line_dict['dihedrals'], dihedral_parser)
    impr_dict = simple_collector(line_dict['improper'], improper_parser)
    
    cmap_lines = {}
    for l in line_dict['cmap']:
        if not (l[0].isdigit() or l[0] == '-'):
            key, n = cmap_key_parser(l)
            cmap_lines[key] = {'n':n, 'entries':[]}
            cur_cmap_lines = cmap_lines[key]['entries']
        else:
            cur_cmap_lines.append(l)

    cmap_dict = {}
    for k, v in cmap_lines.items():
        n, entry_lines = v.values()
        deg_range = range(-180, 180, 360//n)
        id_range = range(0,n**2+1,n)
        entries = cmap_field_parser(entry_lines, n) 
        cmap_dict[k] = []
        for deg, st, end in zip(deg_range, id_range[:-1], id_range[1:]):
            cmap_dict[k].append(cmap_par(deg, entries[st:end]))

    nonbond_dict, nonbond14_dict = extended_collector(
        line_dict['nonbonded'], nonbonded_parser
    )

    nbfix_dict = simple_collector(line_dict['nbfix'], nbfix_parser)
        
    return {
        'bonds': bond_dict,
        'angles': angle_dict,
        'urey_bradley': urey_bradley_dict,
        'dihedrals': dihe_dict,
        'improper': impr_dict,
        'cmap': cmap_dict,
        'nonbonded': nonbond_dict,
        'nonbonded14': nonbond14_dict,
        'nbfix': nbfix_dict
    }


# dictionary store PDBx connect records details
connect_types = {
    'covale': 'covalent bond', 
    'covale_base': 'covalent modification of a nucleotide base',
    'covale_phosphate': 'covalent modification of a nucleotide phosphate',
    'covale_sugar': 'covalent modification of a nucleotide sugar',
    'disulf': 'disulfide bridge',
    'hydrog': 'hydrogen bond',
    'metalc': 'metal coordination',
    'mismat': 'mismatched base pairs', 
    'modres': 'covalent residue modification',
    'saltbr': 'ionic interaction'
}


import warnings
from openbabel import openbabel as ob
from crimm.Fetchers import fetch_rcsb_as_dict
from crimm.Data.ptable import PERIODIC_TABLE

bond_order_dict = {
    # PDB mmCIF bond type names
    'SING': 1,
    'DOUB': 2,
    'TRIP': 3,
    'QUAD': 4,
    'AROM': 2, # we treat it as order of 2 instead of 1.5 here
}

def fetch_bond_info(resname):
    cifdict = fetch_rcsb_as_dict(resname)

    atom_ids = cifdict['chem_comp_atom']['atom_id']
    element_names = cifdict['chem_comp_atom']['type_symbol']
    charges = cifdict['chem_comp_atom']['charge']
    element_dict = {}
    for i, (atom_id, element_name, charge) in enumerate(
        zip(atom_ids, element_names, charges)
    ):
        element_dict[atom_id] = (i, element_name, charge)

    bond_info = cifdict['chem_comp_bond']
    a1, a2 = bond_info['atom_id_1'], bond_info['atom_id_2']
    bond_order = bond_info['value_order']
    aro_flag = [flag == 'Y' for flag in bond_info['pdbx_aromatic_flag']]
    bond_list = list(zip(a1, a2, bond_order, aro_flag))
    return bond_list

def heterogen_to_openbabel(lig):
    bond_list = fetch_bond_info(lig.resname)
    atom_ids = {}
    mol = ob.OBMol()
    mol.SetTitle(lig.resname)
    res = mol.NewResidue()
    # mol.AddResidue(res)
    mol.BeginModify()
    for i, atom in enumerate(lig, start=1):
        atom_ids[atom.name] = i
        atomic_num = PERIODIC_TABLE[atom.element]['number']
        a = mol.NewAtom()
        a.SetAtomicNum(atomic_num)
        a.SetVector(*atom.coord) # coordinates
        ob.OBAtomAssignTypicalImplicitHydrogens(a)

    arom_atoms = set()
    for a1, a2, bond_type, is_arom in bond_list:
        if a1.startswith('H') or a2.startswith('H'):
            # we add hydrogen later according to the user supplied pH value
            continue
        a1_idx, a2_idx = atom_ids[a1], atom_ids[a2]
        bond_order = bond_order_dict[bond_type]
        # AddBond uses 1-indexed atom idx
        ob_atom1, ob_atom2 = mol.GetAtom(a1_idx), mol.GetAtom(a2_idx)
        a1_hs = ob_atom1.GetImplicitHCount()-bond_order
        a2_hs = ob_atom2.GetImplicitHCount()-bond_order
        # floor to zero to avoid negtive number
        a1_hs = (abs(a1_hs)+(a1_hs))//2
        a2_hs = (abs(a2_hs)+(a2_hs))//2
        ob_atom1.SetImplicitHCount(a1_hs)
        ob_atom2.SetImplicitHCount(a2_hs)
        if is_arom:
            arom_atoms.add(a1_idx)
            arom_atoms.add(a2_idx)


    for atom_id in arom_atoms:
        atom = mol.GetAtom(atom_id)
        atom.SetAromatic(True)

    for atom_name, atom_idx in atom_ids.items():
        ob_atom = mol.GetAtom(atom_idx)
        res.AddAtom(ob_atom)
        res.SetAtomID(ob_atom, atom_name)

    mol.SetAromaticPerceived()
    mol.EndModify()

    atom_names = {v:k for k, v in atom_ids.items()}

    # Args: polaronly = false, correctForPH = false, pH = 7.4
    mol.AddHydrogens(False, True, 7.4)

    # Build hydrogen coords with SD minimization using MMFF94
    ff = ob.OBForceField.FindType('mmff94')
    success = ff.Setup(mol)
    if not success:
        warnings.warn(
            'Openbabel MMFF94 forcefield setup failed for hydrogen coord minimization!'
        )
        return
    # Constraints has to be initiated after the forcefield setup
    constraints = ob.OBFFConstraints()
    for atom in ob.OBMolAtomIter(mol):
        res.AddAtom(atom)
        atom_idx = atom.GetIdx() # 1-indexed
        if atom.GetAtomicNum() == 1:
            res.SetAtomID(atom, 'H'+str(atom_idx))
            atom.ClearCoordPtr()
        else:
            res.SetAtomID(atom, atom_names[atom_idx])
            # AddAtomConstraint uses 0-indexed atom id
            constraints.AddAtomConstraint(atom.GetId())
    res.SetName(lig.resname)
    res.SetNum(lig.id[1])
    ff.SetConstraints(constraints)
    ff.SteepestDescent(100)
    ff.GetCoordinates(mol)
    return mol



"""
Module for writing CHARMM CARD coordinate files (CRD).

This module provides functions to write coordinate data from crimm entities
to CHARMM CRD format files. The CRD format stores atomic coordinates along
with residue and segment information.

Format specification (from CHARMM source io/coorio.F90):
- Extended format: 2I10, 2X,A8, 2X,A8, 3F20.10, 2X,A8, 2X,A8, F20.10
- Standard format: 2I5, 1X,A4, 1X,A4, 3F10.5, 1X,A4, 1X,A4, F10.5
"""

from typing import Union, List
from crimm.StructEntities.Atom import Atom
from crimm.StructEntities.Residue import Residue
from crimm.StructEntities.Chain import Chain
from crimm.StructEntities.Model import Model
from crimm.StructEntities.Structure import Structure


# Extended format (EXT) - for systems with >99999 atoms
_EXT_HEADER_FORMAT = "{:>10d}  EXT\n"
_EXT_ATOM_FORMAT = (
    "{serial:>10d}{resno:>10d}  {resname:<8s}  {atomname:<8s}"
    "{x:>20.10f}{y:>20.10f}{z:>20.10f}  {segid:<8s}  {resid:<8s}{weight:>20.10f}\n"
)

# Standard format - for smaller systems
_STD_HEADER_FORMAT = "{:>5d}\n"
_STD_ATOM_FORMAT = (
    "{serial:>5d}{resno:>5d} {resname:<4s} {atomname:<4s}"
    "{x:>10.5f}{y:>10.5f}{z:>10.5f} {segid:<4s} {resid:<4s}{weight:>10.5f}\n"
)


def _get_atoms_from_entity(entity, include_lonepairs: bool = True) -> List[Atom]:
    """Get all atoms from an entity, optionally including lone pairs.

    Parameters
    ----------
    entity : Model, Structure, Chain, or Residue
        The entity to extract atoms from
    include_lonepairs : bool
        Whether to include lone pair pseudo-atoms for CGENFF ligands

    Returns
    -------
    List[Atom]
        List of atoms in order
    """
    atoms = []

    if hasattr(entity, 'get_atoms'):
        atoms.extend(list(entity.get_atoms()))
    elif isinstance(entity, Atom):
        atoms.append(entity)

    # Include lone pairs if present and requested
    if include_lonepairs:
        if hasattr(entity, 'get_residues'):
            for residue in entity.get_residues():
                if hasattr(residue, 'lone_pair_dict') and residue.lone_pair_dict:
                    atoms.extend(residue.lone_pair_dict.values())
        elif hasattr(entity, 'lone_pair_dict') and entity.lone_pair_dict:
            atoms.extend(entity.lone_pair_dict.values())

    return atoms


def _generate_system_info(entity) -> list:
    """Generate informative title lines from entity metadata.

    Extracts useful information about the system:
    - Structure/PDB ID
    - Chain IDs and types
    - Disulfide bonds
    - CGenFF usage for ligands
    - Total atom count

    Parameters
    ----------
    entity : Model, Structure, Chain, or Residue
        The entity to extract information from

    Returns
    -------
    list
        List of title lines describing the system
    """
    import datetime
    import getpass

    lines = []

    # Get timestamp and user
    timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    try:
        user = getpass.getuser()
    except Exception:
        user = "unknown"

    # Get structure ID (PDB ID)
    structure_id = None
    model = None

    if hasattr(entity, 'level'):
        if entity.level == 'M':  # Model
            model = entity
            if hasattr(entity, 'parent') and entity.parent is not None:
                structure_id = entity.parent.id
        elif entity.level == 'S':  # Structure
            structure_id = entity.id
            models = list(entity.get_models())
            if models:
                model = models[0]
        elif entity.level == 'C':  # Chain
            model = entity.parent
            if model and hasattr(model, 'parent') and model.parent:
                structure_id = model.parent.id

    # Build first line
    if structure_id:
        lines.append(f"Structure: {structure_id} | Generated by crimm")
    else:
        lines.append("Generated by crimm")

    # Get chain info
    chain_info = []
    disu_count = 0
    has_cgenff = False
    has_tip3 = False
    ion_types = set()

    if model is not None:
        for chain in model:
            chain_id = chain.id
            chain_type = getattr(chain, 'chain_type', 'Unknown')

            # Count disulfide bonds from patches
            if hasattr(chain, 'topology') and chain.topology is not None:
                topo = chain.topology
                if hasattr(topo, 'inter_res_patches'):
                    for patch in topo.inter_res_patches:
                        if hasattr(patch, 'name') and 'DISU' in patch.name.upper():
                            disu_count += 1
                # Check for CGenFF
                if hasattr(topo, 'source') and topo.source:
                    if 'cgenff' in str(topo.source).lower():
                        has_cgenff = True

            # Categorize chains
            if chain_type in ('Polypeptide(L)', 'Protein'):
                n_res = len(list(chain.get_residues()))
                chain_info.append(f"{chain_id}(protein,{n_res}res)")
            elif chain_type == 'Solvent':
                has_tip3 = True
            elif chain_type == 'Ion':
                for res in chain.get_residues():
                    ion_types.add(res.resname.strip())
            elif chain_type == 'Ligand':
                chain_info.append(f"{chain_id}(ligand)")
            elif chain_type in ('Polyribonucleotide', 'RNA'):
                chain_info.append(f"{chain_id}(RNA)")
            elif chain_type in ('Polydeoxyribonucleotide', 'DNA'):
                chain_info.append(f"{chain_id}(DNA)")

    # Add chain summary
    if chain_info:
        lines.append(f"Chains: {', '.join(chain_info)}")

    # Count waters and ions
    n_waters = 0
    n_ions = 0
    if model is not None:
        for chain in model:
            chain_type = getattr(chain, 'chain_type', 'Unknown')
            if chain_type == 'Solvent':
                n_waters = len(list(chain.get_residues()))
            elif chain_type == 'Ion':
                n_ions = len(list(chain.get_residues()))

    # Check for solvation metadata (salt concentration, preserved waters)
    solvation_info = []
    if model is not None and hasattr(model, '_solvation_info'):
        sol_info = model._solvation_info
        if 'concentration' in sol_info:
            solvation_info.append(f"{sol_info['concentration']*1000:.0f} mM salt")
        if sol_info.get('preserved_waters', 0) > 0:
            solvation_info.append(f"{sol_info['preserved_waters']} crystal waters kept")
        if sol_info.get('preserved_ions', 0) > 0:
            solvation_info.append(f"{sol_info['preserved_ions']} crystal ions kept")

    # Add solvation summary
    if n_waters > 0 or n_ions > 0:
        solv_parts = []
        if n_waters > 0:
            solv_parts.append(f"{n_waters} waters")
        if n_ions > 0:
            solv_parts.append(f"{n_ions} ions ({','.join(sorted(ion_types))})")
        if solvation_info:
            solv_parts.extend(solvation_info)
        lines.append(f"Solvation: {', '.join(solv_parts)}")

    # Add crystal/box information
    if model is not None and hasattr(model, '_solvation_info'):
        sol_info = model._solvation_info
        box_type = sol_info.get('box_type')
        box_dim = sol_info.get('box_dim')
        if box_type and box_dim:
            box_type_names = {
                'cube': 'Cubic',
                'octa': 'Truncated Octahedron',
                'rhdo': 'Rhombic Dodecahedron',
                'ortho': 'Orthorhombic',
                'tetra': 'Tetragonal',
                'hexa': 'Hexagonal',
                'mono': 'Monoclinic',
                'tric': 'Triclinic',
                'rhomb': 'Rhombohedral',
            }
            box_name = box_type_names.get(box_type, box_type.capitalize())
            # Handle orthorhombic with different a, b, c dimensions
            box_dims = sol_info.get('box_dims')
            if box_type == 'ortho' and box_dims is not None:
                a, b, c = box_dims
                lines.append(f"Crystal: {box_name} box, {a:.2f} x {b:.2f} x {c:.2f} A")
            else:
                lines.append(f"Crystal: {box_name} box, {box_dim:.2f} A")

    # Add special features
    features = []
    if disu_count > 0:
        features.append(f"{disu_count} DISU")
    if has_cgenff:
        features.append("CGenFF")

    if features:
        lines.append(f"Features: {' | '.join(features)}")

    # Add timestamp
    lines.append(f"Created: {timestamp} by {user}")

    return lines


def _format_title_lines(title: str = "", entity=None) -> str:
    """Format title lines with asterisk prefix.

    CHARMM CRD format requires title section to have at least two lines:
    one or more content lines starting with "* " followed by a terminating "*".

    Parameters
    ----------
    title : str, optional
        Title text (can be multiline). If empty and entity is provided,
        system info is extracted from the entity. Otherwise a basic
        default title is used.
    entity : Model, Structure, Chain, or Residue, optional
        Entity to extract system information from for auto-generated title.

    Returns
    -------
    str
        Formatted title lines with asterisk prefix
    """
    import datetime
    import getpass

    lines = []
    if title:
        # User-provided title
        for line in title.strip().split('\n'):
            lines.append(f"* {line}\n")
    elif entity is not None:
        # Auto-generate informative title from entity
        for line in _generate_system_info(entity):
            lines.append(f"* {line}\n")
    else:
        # Basic default title
        timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        try:
            user = getpass.getuser()
        except Exception:
            user = "unknown"
        lines.append(f"* CRD file generated by crimm\n")
        lines.append(f"* Created: {timestamp} by {user}\n")
    # Terminating asterisk (required by CHARMM format)
    lines.append("*\n")
    return "".join(lines)


def get_crd_str(
    entity: Union[Model, Structure, Chain, Residue],
    extended: bool = True,
    title: str = "",
    include_lonepairs: bool = True,
    reset_serial: bool = True
) -> str:
    """Get CHARMM CRD format string from a crimm entity.

    Parameters
    ----------
    entity : Model, Structure, Chain, or Residue
        The entity to write coordinates from
    extended : bool, default True
        Use extended format (I10/A8/F20.10) for large systems
    title : str, default ""
        Title line(s) to include in the header
    include_lonepairs : bool, default True
        Include lone pair pseudo-atoms for CGENFF ligands
    reset_serial : bool, default True
        Reset atom serial numbers starting from 1

    Returns
    -------
    str
        CRD format string
    """
    # Get atoms
    atoms = _get_atoms_from_entity(entity, include_lonepairs)
    natom = len(atoms)

    if natom == 0:
        raise ValueError("No atoms found in entity")

    # Select format based on extended flag and atom count
    if extended or natom >= 100000:
        header_fmt = _EXT_HEADER_FORMAT
        atom_fmt = _EXT_ATOM_FORMAT
        str_width = 8
    else:
        header_fmt = _STD_HEADER_FORMAT
        atom_fmt = _STD_ATOM_FORMAT
        str_width = 4

    # Build output
    lines = []

    # Title section (pass entity for auto-generated system info)
    lines.append(_format_title_lines(title, entity=entity))

    # Atom count header
    lines.append(header_fmt.format(natom))

    # Build global residue numbering map
    # RESNO in CRD format must be sequential across the entire system (1, 2, 3...)
    # while RESID is the segment-local residue identifier
    resno_map = {}  # Maps residue object to global resno
    global_resno = 1
    last_residue = None

    for atom in atoms:
        residue = atom.parent
        if residue is not last_residue and residue is not None:
            if residue not in resno_map:
                resno_map[residue] = global_resno
                global_resno += 1
            last_residue = residue

    # Atom lines
    for idx, atom in enumerate(atoms, start=1):
        serial = idx if reset_serial else (atom.serial_number or idx)

        # Get parent residue info
        residue = atom.parent
        if residue is not None:
            resname = residue.resname[:str_width]
            if resname == 'HIS' and residue.topo_definition is not None:
                # Use residue definition resname for HIS variants
                resname = residue.topo_definition.resname
            segid = (residue.segid or "")[:str_width]
            # Use global sequential RESNO for CHARMM compatibility
            resno = resno_map.get(residue, 1)
            # RESID is the segment-local residue ID as string
            resid = str(residue.id[1])[:str_width]
        else:
            resname = "UNK"[:str_width]
            segid = ""[:str_width]
            resno = 1
            resid = "1"[:str_width]

        atomname = atom.name[:str_width]

        # Get coordinates
        if atom.coord is None:
            raise ValueError(f"Atom {atom.name} (serial {serial}) has no coordinates")
        x, y, z = atom.coord

        # Validate coordinates are finite numbers
        import numpy as np
        if not all(np.isfinite([x, y, z])):
            raise ValueError(
                f"Atom {atom.name} (serial {serial}) has invalid coordinates: "
                f"({x}, {y}, {z}). Coordinates must be finite numbers."
            )

        # Weight/temperature factor
        weight = atom.bfactor if atom.bfactor is not None else 0.0

        lines.append(atom_fmt.format(
            serial=serial,
            resno=resno,
            resname=resname,
            atomname=atomname,
            x=x,
            y=y,
            z=z,
            segid=segid,
            resid=resid,
            weight=weight
        ))

    return "".join(lines)


def write_crd(
    entity: Union[Model, Structure, Chain, Residue],
    filepath: str,
    extended: bool = True,
    title: str = "",
    include_lonepairs: bool = True,
    reset_serial: bool = True
) -> None:
    """Write CHARMM CRD format file from a crimm entity.

    Parameters
    ----------
    entity : Model, Structure, Chain, or Residue
        The entity to write coordinates from
    filepath : str
        Path to output file
    extended : bool, default True
        Use extended format (I10/A8/F20.10) for large systems
    title : str, default ""
        Title line(s) to include in the header
    include_lonepairs : bool, default True
        Include lone pair pseudo-atoms for CGENFF ligands
    reset_serial : bool, default True
        Reset atom serial numbers starting from 1
    """
    crd_str = get_crd_str(
        entity,
        extended=extended,
        title=title,
        include_lonepairs=include_lonepairs,
        reset_serial=reset_serial
    )

    with open(filepath, 'w', encoding='utf-8') as f:
        f.write(crd_str)



# Code for Sybyl atom and bond typing is from Open Drug Discovery Toolkit (ODDT)
# https://github.com/oddt/oddt/blob/master/oddt/toolkits/extras/rdkit/__init__.py
# L151
# Copyright (c) 2014, Maciej Wójcikowski
# Copyright (c) 2023, Truman Xu (徐梓乔), Brooks Lab at the University of Michigan
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of ODDT or crimm nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import warnings
import requests, json
from crimm.Fetchers import fetch_rcsb_as_dict
from crimm.IO.PDBString import get_pdb_str
from crimm.StructEntities.Atom import Atom
from crimm.StructEntities.Residue import Heterogen
import numpy as np
from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import BondType as rdBond
from rdkit.Geometry import Point3D
from crimm.Data.components_dict import TRANSITION_METALS, ALL_METALS

bond_order_dict = {
    # PDB mmCIF bond type names
    'SING': rdBond.SINGLE,
    'DOUB': rdBond.DOUBLE,
    'TRIP': rdBond.TRIPLE,
    'QUAD': rdBond.QUADRUPLE,
    'AROM': rdBond.AROMATIC,
    # probe bond type names
    'single': rdBond.SINGLE,
    'double': rdBond.DOUBLE,
    'triple': rdBond.TRIPLE,
    'quadruple': rdBond.QUADRUPLE,
    'aromatic': rdBond.AROMATIC,
}

def get_rdkit_bond_order(bo_name):
    return bond_order_dict.get(bo_name, rdBond.OTHER)

####################### ODDT code starts #######################
# Mol2 Atom typing
def _sybyl_atom_type(atom):
    """ Asign sybyl atom type
    Reference #1: http://www.tripos.com/mol2/atom_types.html
    Reference #2: http://chemyang.ccnu.edu.cn/ccb/server/AIMMS/mol2.pdf
    """
    sybyl = None
    atom_symbol = atom.GetSymbol()
    atomic_num = atom.GetAtomicNum()
    hyb = atom.GetHybridization()-1  # -1 since 1 = sp, 2 = sp1 etc
    hyb = min(hyb, 3)
    degree = atom.GetDegree()
    aromtic = atom.GetIsAromatic()

    # define groups for atom types
    guanidine = '[NX3,NX2]([!O,!S])!@C(!@[NX3,NX2]([!O,!S]))!@[NX3,NX2]([!O,!S])'  # strict
    # guanidine = '[NX3]([!O])([!O])!:C!:[NX3]([!O])([!O])' # corina compatible
    # guanidine = '[NX3]!@C(!@[NX3])!@[NX3,NX2]'
    # guanidine = '[NX3]C([NX3])=[NX2]'
    # guanidine = '[NX3H1,NX2,NX3H2]C(=[NH1])[NH2]' # previous
    #

    if atomic_num == 6:
        if aromtic:
            sybyl = 'C.ar'
        elif degree == 3 and _atom_matches_smarts(atom, guanidine):
            sybyl = 'C.cat'
        else:
            sybyl = '%s.%i' % (atom_symbol, hyb)
    elif atomic_num == 7:
        if aromtic:
            sybyl = 'N.ar'
        elif _atom_matches_smarts(atom, 'C(=[O,S])-N'):
            sybyl = 'N.am'
        elif degree == 3 and _atom_matches_smarts(atom, '[$(N!-*),$([NX3H1]-*!-*)]'):
            sybyl = 'N.pl3'
        elif _atom_matches_smarts(atom, guanidine):  # guanidine has N.pl3
            sybyl = 'N.pl3'
        elif degree == 4 or hyb == 3 and atom.GetFormalCharge():
            sybyl = 'N.4'
        else:
            sybyl = '%s.%i' % (atom_symbol, hyb)
    elif atomic_num == 8:
        # http://www.daylight.com/dayhtml_tutorials/languages/smarts/smarts_examples.html
        if degree == 1 and _atom_matches_smarts(atom, '[CX3](=O)[OX1H0-]'):
            sybyl = 'O.co2'
        elif degree == 2 and not aromtic:  # Aromatic Os are sp2
            sybyl = 'O.3'
        else:
            sybyl = 'O.2'
    elif atomic_num == 16:
        # http://www.daylight.com/dayhtml_tutorials/languages/smarts/smarts_examples.html
        if degree == 3 and _atom_matches_smarts(atom, '[$([#16X3]=[OX1]),$([#16X3+][OX1-])]'):
            sybyl = 'S.O'
        # https://github.com/rdkit/rdkit/blob/master/Data/FragmentDescriptors.csv
        elif _atom_matches_smarts(atom, 'S(=,-[OX1;+0,-1])(=,-[OX1;+0,-1])(-[#6])-[#6]'):
            sybyl = 'S.o2'
        else:
            sybyl = '%s.%i' % (atom_symbol, hyb)
    elif atomic_num == 15 and hyb == 3:
        sybyl = '%s.%i' % (atom_symbol, hyb)

    if not sybyl:
        sybyl = atom_symbol
    return sybyl


def _atom_matches_smarts(atom, smarts):
    idx = atom.GetIdx()
    patt = Chem.MolFromSmarts(smarts)
    for m in atom.GetOwningMol().GetSubstructMatches(patt):
        if idx in m:
            return True
    return False

def _amide_bond(bond):
    a1 = bond.GetBeginAtom()
    a2 = bond.GetEndAtom()
    if (a1.GetAtomicNum() == 6 and a2.GetAtomicNum() == 7 or
            a2.GetAtomicNum() == 6 and a1.GetAtomicNum() == 7):
        # https://github.com/rdkit/rdkit/blob/master/Data/FragmentDescriptors.csv
        patt = Chem.MolFromSmarts('C(=O)-N')
        for m in bond.GetOwningMol().GetSubstructMatches(patt):
            if a1.GetIdx() in m and a2.GetIdx() in m:
                return True
    return False

######################## ODDT code ends ########################

tripos_bond_order_dict = {
    rdBond.SINGLE : 1,
    rdBond.DOUBLE : 2,
    rdBond.TRIPLE : 3,
    rdBond.AROMATIC : 'ar',
}

def _check_amide_bond(rdk_bond):
    if _amide_bond(rdk_bond):
        return 'am'
    return 'un'

def determine_tripos_bond_type(rdk_bond):
    return tripos_bond_order_dict.get(
        rdk_bond.GetBondType(), _check_amide_bond(rdk_bond)
    )

def _generate_mol2_title_block(resname, n_atoms, n_bonds):
    title_block = f"""# Generated by crimm with rdkit
@<TRIPOS>MOLECULE
{resname}
{n_atoms} {n_bonds} 1
SMALL
GASTEIGER
****

"""
    return title_block

def _determine_ligname(mol):
    atom = list(mol.GetAtoms())[0]
    if info := atom.GetPDBResidueInfo():
        ligname = info.GetResidueName()
    elif mol.HasProp('Description'):
        ligname = mol.GetProp('Description')
    elif mol.HasProp('_Name'):
        ligname = mol.GetProp('_Name')
    else:
        ligname = 'LIG'
    return ligname.upper()[:3]

def _generate_mol2_atom_name(atom, atom_counts: dict):
    """Generate the mol2 atom name for the atom."""
    symbol = atom.GetSymbol()
    if symbol not in atom_counts:
        atom_counts[symbol] = 1
    else:
        atom_counts[symbol] += 1
    
    # PDB-style naming (Element + Number)
    atom_name = f"{symbol}{atom_counts[symbol]}"
    return atom_name

def MolToMol2Block(mol, ligname = None):
    """Write a mol2 block string from a RDKit Mol Object."""
    if ligname is None:
        ligname = _determine_ligname(mol)

    if len(mol.GetConformers()) == 0:
        # No conformer generated, use 2D mol coords
        AllChem.Compute2DCoords(mol)

    pos = mol.GetConformer().GetPositions()
    atoms = list(mol.GetAtoms())
    if not atoms[0].HasProp('_GasteigerCharge'):
        Chem.rdPartialCharges.ComputeGasteigerCharges(mol)

    tripos_atom_format = '{:<6} {:<5} {:>7.3f} {:>7.3f} {:>7.3f} {:<6} {} {} {:>6.3f}'
    tripos_bond_format = '{:<5} {:<5} {:<5} {:<2}'
    atom_lines = []
    bond_lines = []
    atom_counts={}
    for i, (atom, coords) in enumerate(zip(atoms, pos), start=1):
        if pdbinfo := atom.GetPDBResidueInfo():
            name = pdbinfo.GetName()
        else:
            name = _generate_mol2_atom_name(atom, atom_counts)
        sybyl_type = _sybyl_atom_type(atom)
        charge = atom.GetDoubleProp('_GasteigerCharge')
        atom_lines.append(
            tripos_atom_format.format(
                i, name, *coords, sybyl_type, 1, ligname, charge
            )
        )
        

    tripos_bonds = []
    for bond in mol.GetBonds():
        st = bond.GetBeginAtomIdx()+1
        end = bond.GetEndAtomIdx()+1
        ordered_ids = sorted((st, end))
        bo = determine_tripos_bond_type(bond)
        tripos_bonds.append((*ordered_ids, bo))

    for i, bond_info in enumerate(sorted(tripos_bonds), start=1):
        bond_lines.append(tripos_bond_format.format(i, *bond_info))

    n_atoms, n_bonds = len(atom_lines), len(bond_lines)
    block_string = ""
    block_string += _generate_mol2_title_block(ligname, n_atoms, n_bonds)
    block_string += '@<TRIPOS>ATOM\n'
    for l in atom_lines:
        block_string += l+'\n'
    block_string += '@<TRIPOS>BOND\n'
    for l in bond_lines:
        block_string += l+'\n'
    block_string += f'@<TRIPOS>SUBSTRUCTURE\n1 {ligname} 1'
    return block_string

def MolToMol2File(mol, ligname = None, filename = None):
    """Write a mol2 file from a RDKit Mol Object."""
    mol2_string = MolToMol2Block(mol=mol, ligname=ligname)

    if filename is None:
        filename = f'{ligname}.mol2'
    with open(filename, 'w', encoding='UTF-8') as f:
        f.write(mol2_string)

class LigandBondOrderException(Exception):
    """Define class LigandBondOrderException."""

class SmilesQueryException(Exception):
    """Define class SmilesQueryException."""

class LigandBondOrderWarning(Warning):
    """Define class LigandBondOrderException."""

class SmilesQueryWarning(Warning):
    """Define class SmilesQueryException."""

class RDKitHetConverter:
    """Convert a crimm heterogen object to an rdkit mol object. The mol object 
    WILL have the correct bond orders. Upon loading the heterogen, the CIF 
    definition is queried from rcsb.org, which contains all the necessary
    information to create the rdkit mol object. The mol object is then used to
    generate a mol2 file.

    Attributes:
        lig (Heterogen): The heterogen object to be converted.
        lig_pdbid (str): The PDB ID of the heterogen.
        chain_id (str): The chain ID of the heterogen.
        element_dict (dict): A dictionary mapping the atom name to the element
            symbol.
        bond_dict (list): A list of tuples containing the bond information.
        resname (str): The residue name of the heterogen.
        resnum (int): The residue number of the heterogen.
        mol (rdkit.Chem.rdchem.Mol): The constructed rdkit mol object.
    
    Raises:
        TypeError: If the input is not a crimm heterogen object constructed from
        pdb mmcif file.
    
    Methods:
        load_heterogen: Load a heterogen object and create the rdkit mol object.
        get_mol: Return the rdkit mol object.
        write_mol2: Write the mol2 file of the loaded heterogen.
    
    Examples:
        >>> from crimm import fetch_rcsb
        >>> from crimm.Adaptors import RDKitHetConverter
        >>> model = fetch_rcsb('1aka')
        ### Select the heterogen from the structure
        ### Note that the heterogen is at residue level not chain level
        >>> lig = model['D'].residues[0]
        >>> rdk_converter = RDKitHetConverter()
        >>> rdk_converter.load_heterogen(lig)
        >>> mol = rdk_converter.get_mol()
        >>> rdk_converter.write_mol2()
    """
    def __init__(self):
        self.lig = None
        self.lig_pdbid = None
        self.chain_id = None
        self.element_dict = None
        self.bond_dict = None
        self.resname = None
        self.resnum = None
        self._edmol = Chem.EditableMol(Chem.Mol())
        self._pending_hydrogens = None
        self._rd_atoms = None
        self._sanitize = True # determine if sanitization can be performed
        self.mol = None
        self._atom_counts = None

    def load_heterogen(self, lig, update_hydrogens = True):
        """Load a heterogen object and create the rdkit mol object. If the
        heterogen contains metal atoms, the mol object will not be sanitized.
        Hydrogens will not be added to the mol object if the heterogen contains
        metal atoms. If the heterogen does not contain metal atoms, the mol
        object will be sanitized and hydrogens will be added to the mol object.
        Args:
            lig (Heterogen): The heterogen residue object to be converted.
            update_hydrogens (bool): If True, hydrogens positions calculted by rdkit
                will be updated on the crimm ligand residue. Defaults to True.
        """
        if isinstance(lig, Heterogen):
            self.lig = lig
        else:
            raise TypeError("Input must be a Heterogen object (Residue level).")

        self.lig_pdbid = lig.resname
        self.chain_id = lig.parent.id
        self.resname = lig.resname
        self.resnum = lig.id[1]
        self._edmol = Chem.EditableMol(Chem.Mol())
        self._pending_hydrogens = None
        self._rd_atoms = None
        self._sanitize = True
        self.mol = None

        cifdict = fetch_rcsb_as_dict(self.lig_pdbid)

        atom_ids = cifdict['chem_comp_atom']['atom_id']
        element_names = cifdict['chem_comp_atom']['type_symbol']
        charges = cifdict['chem_comp_atom']['charge']
        self.element_dict = {}
        for i, (atom_id, element_name, charge) in enumerate(
            zip(atom_ids, element_names, charges)
        ):
            if element_name.capitalize() in ALL_METALS:
                # if the heterogen contains a metal, we do not perform sanitization
                self._sanitize = False
            self.element_dict[atom_id] = (i, element_name, charge)

        bond_info = cifdict['chem_comp_bond']
        a1, a2 = bond_info['atom_id_1'], bond_info['atom_id_2']
        bond_order = bond_info['value_order']
        aro_flag = [flag == 'Y' for flag in bond_info['pdbx_aromatic_flag']]
        self.bond_dict = list(zip(a1, a2, bond_order, aro_flag))
        self._create_rdkit_mol()
        if update_hydrogens:
            self.update_hydrogens()
    
    def load_rdkit_mol(self, mol, ligand_resname = None):
        """Load a rdkit mol object and create the crimm heterogen object. If the
        mol is 2D, a 3D conformer will be generated. The heterogen object will
        be created with the atom information from the mol object. The bond
        information will be stored in the topo_definition attribute of the
        heterogen object.
        
        Args:
            mol (rdkit.Chem.rdchem.Mol): The rdkit mol object to be converted.
        """
        self.mol = mol
        
        if ligand_resname is None:
            self.lig_pdbid = _determine_ligname(mol)
        else:
            self.lig_pdbid = ligand_resname
        self.resname = self.lig_pdbid
        self.resnum = 1
        res_id = (f'H_{self.resname}', self.resnum, ' ')
        self.lig = Heterogen(res_id, self.resname, ' ')
        self._sanitize = True

        if len(mol.GetConformers()) == 0:
            # Generate conformer if not present
            params = AllChem.ETKDGv3()
            AllChem.EmbedMolecule(mol, params)

        pos = mol.GetConformer().GetPositions()
        atoms = list(mol.GetAtoms())
        

        atom_counts = {}
        atom_mapping = {}
        for i, (atom, coords) in enumerate(zip(atoms, pos), start=1):
            atom_id = atom.GetIdx()
            if pdbinfo := atom.GetPDBResidueInfo():
                name = pdbinfo.GetName()
            else:
                name = _generate_mol2_atom_name(atom, atom_counts)
                pdb_info = Chem.AtomPDBResidueInfo(
                    name, i, ' ', 
                    residueName = self.resname,
                    residueNumber = self.resnum,
                    isHeteroAtom = True
                )
                atom.SetPDBResidueInfo(pdb_info)
            name = name.strip()
            new_atom = Atom(
                name = name,
                coord=coords,
                bfactor=0.0,
                occupancy=1.0,
                altloc=' ',
                serial_number=i,
                element=atom.GetSymbol().upper(),
                fullname=f"{name:^4}",
                topo_definition=None
            )
            self.lig.add(new_atom)
            atom_mapping[atom_id] = name

        bond_dict = {}
        for bond in mol.GetBonds():
            st = atom_mapping[bond.GetBeginAtomIdx()]
            end = atom_mapping[bond.GetEndAtomIdx()]

            bo_name = str(bond.GetBondType()).lower()
            bond_dict[bo_name] = (st, end)

        self.lig.bonds = bond_dict
        self.lig._rdkit_mol = mol

    def _create_rdkit_PDBinfo(self, pdb_atom_name, altloc):
        serial_number = self.element_dict[pdb_atom_name][0]
        pdb_info = Chem.AtomPDBResidueInfo(
            pdb_atom_name, serial_number, altloc, residueName = self.resname,
            residueNumber = self.resnum, chainId = self.chain_id,
            isHeteroAtom = True
        )
        return pdb_info
    
    def _create_rdkit_atoms(self):
        self._rd_atoms = {}
        for atom in self.lig.atoms:
            rd_atom = Chem.Atom(atom.element.capitalize())
            pdb_info = self._create_rdkit_PDBinfo(atom.name, atom.altloc)
            rd_atom.SetPDBResidueInfo(pdb_info)
            charge = self.element_dict[atom.name][2]
            element = self.element_dict[atom.name][1].capitalize()
            rd_atom.SetFormalCharge(charge)
            atom_idx = self._edmol.AddAtom(rd_atom)
            self._rd_atoms[atom.name] = atom_idx

    def _add_rdkit_bonds(self):
        self._pending_hydrogens = {}
        for a1, a2, bo_name, is_arom in self.bond_dict:
            elem1 = self.element_dict[a1][1].capitalize()
            elem2 = self.element_dict[a2][1].capitalize()
            n_transition_metals = (
                elem1 in TRANSITION_METALS
            ) + (
                elem2 in TRANSITION_METALS
            )
            if a1 not in self._rd_atoms:
                if a2 not in self._pending_hydrogens:
                    self._pending_hydrogens[a2] = []
                self._pending_hydrogens[a2].append(a1)
                if self.element_dict[a1][1] != 'H':
                    raise ValueError(
                        'CIF definition does not match the heterogen instance! '
                        f'Fail to find atom {a1} in mol {self.resname}'
                    )
                continue
            elif a2 not in self._rd_atoms:
                if a1 not in self._pending_hydrogens:
                    self._pending_hydrogens[a1] = []
                self._pending_hydrogens[a1].append(a2)
                if self.element_dict[a2][1] != 'H':
                    raise ValueError(
                        'CIF definition does not match the heterogen instance! '
                        f'Fail to find atom {a2} in mol {self.resname}'
                    )
                continue
            if n_transition_metals == 1:
                # we assume metal-ligand bond is a dative bond 
                bo = rdBond.DATIVE
            else:
                bo = get_rdkit_bond_order(bo_name)
            idx1, idx2 = self._rd_atoms[a1], self._rd_atoms[a2]
            self._edmol.AddBond(idx1, idx2, bo)

    def _create_conformer(self):
        conf = Chem.Conformer()
        conf.Set3D(True)
        for atom in self.lig:
            idx = self._rd_atoms[atom.name]
            coord = Point3D(*atom.coord)
            conf.SetAtomPosition(idx, coord)
        return conf

    def _add_hydrogen_PDBinfo(self, mol):
        for heavy_atom_name, hydrogen_names in self._pending_hydrogens.items():
            heavy_atom = mol.GetAtomWithIdx(self._rd_atoms[heavy_atom_name])
            nei_hs = []
            for nei in heavy_atom.GetNeighbors():
                if nei.GetAtomicNum() == 1:
                    nei_hs.append(nei)
            for hydrogen_name, atom in zip(hydrogen_names, nei_hs):
                pdb_info = self._create_rdkit_PDBinfo(hydrogen_name, '')
                atom.SetPDBResidueInfo(pdb_info)

    def _create_rdkit_mol(self):
        self._create_rdkit_atoms()
        self._add_rdkit_bonds()
        mol = self._edmol.GetMol()
        conf = self._create_conformer()
        mol.AddConformer(conf)
        if not self._sanitize:
            warnings.warn(
                'The heterogen contains metal atoms! Sanitization is not performed.'
                'Hygrogen atoms are not added to the mol object.'
            )
            self.mol = mol
            return
        AllChem.SanitizeMol(mol)
        mol = AllChem.AddHs(mol, addCoords=True)
        AllChem.SanitizeMol(mol)
        AllChem.ComputeGasteigerCharges(mol)
        self._add_hydrogen_PDBinfo(mol)
        Chem.AssignStereochemistryFrom3D(mol)
        self.mol = mol

    def get_mol(self):
        """Return the rdkit mol object."""
        if self.mol is None:
            self._create_rdkit_mol()
        return self.mol

    def get_ligand(self):
        """Return the crimm heterogen object."""
        return self.lig
    
    def write_mol2(self, filename = None):
        """Write the mol2 file of the loaded heterogen."""
        MolToMol2File(self.mol, self.resname, filename)

    def get_mol2_block(self):
        """Return the mol2 block of the loaded heterogen."""
        return MolToMol2Block(self.mol, self.resname)

    def update_hydrogens(self):
        """Add hydrogens to the ligand residue object."""
        if self.mol is None:
            self._create_rdkit_mol()
        last_serial = list(self.lig.get_atoms())[0].serial_number+1
        for atom, coord in zip(
            self.mol.GetAtoms(), self.mol.GetConformer(0).GetPositions()
        ):
            atom_name = atom.GetPDBResidueInfo().GetName()

            if atom.GetSymbol() == 'H' and atom_name not in self.lig:
                new_atom = Atom(
                    name = atom_name,
                    coord=coord,
                    bfactor=0.0,
                    occupancy=1.0,
                    altloc=' ',
                    serial_number=last_serial,
                    element=atom.GetSymbol(),
                    fullname=f"{atom_name:^4}",
                    topo_definition=None
                )
                self.lig.add(new_atom)
                last_serial+=1

def _build_smiles_query(resname):
    query = '''
    {
        chem_comps(comp_ids: ["{var_lig_id}"]) {
            rcsb_id
            rcsb_chem_comp_descriptor {
            SMILES
            }
        }
    }
    '''.replace("{var_lig_id}", resname)
    return query

def query_rcsb_for_smiles(het_res):
    """Query the canonical SMILES for the heterogen molecule from PDB based on chem_comps
    ID"""
    url="https://data.rcsb.org/graphql"
    response = requests.post(
            url, json={'query': _build_smiles_query(het_res.resname)}, timeout=1000
        )
    response_dict = json.loads(response.text)
    return_vals = response_dict['data']['chem_comps']
    if not return_vals:
        warnings.warn(
            f'Query on {het_res.resname} did not return requested information: '
            'chem_comps (smiles string)'
        )
        return

    smiles = return_vals[0]['rcsb_chem_comp_descriptor']['SMILES']
    return smiles

def heterogen_to_rdkit(het_res, smiles=None):
    """Convert a heterogen to an rdkit mol. The bond orders are set based on the
    SMILES string. If the SMILES string is not provided, the function will query
    the RCSB PDB for the SMILES string.
    
    Args:
        het_res (Heterogen): The heterogen to be converted.
        smiles (str, optional): The SMILES string of the heterogen. Defaults to None.
        
    Returns:
        rdkit.Chem.rdchem.Mol: The rdkit mol of the heterogen.
    """
    mol = Chem.MolFromPDBBlock(get_pdb_str(het_res))
    if smiles is None:
        smiles = query_rcsb_for_smiles(het_res)
    # We do not allow rdkit mol return if the correct bond orders are not set
    if smiles is None:
        msg = (
            'Fail to set bond orders on the Ligand mol! PDB query on SMILES does not'
            'return any result.'
        )
        warnings.warn(msg, SmilesQueryWarning)
        return

    template = AllChem.MolFromSmiles(smiles)
    template = Chem.RemoveHs(template)
    try:
        mol = AllChem.AssignBondOrdersFromTemplate(template, mol)
        if het_res.pdbx_description is not None:
            mol.SetProp('Description', str(het_res.pdbx_description))
        return mol

    except ValueError:
        msg = (
            'No structure match found! Possibly the SMILES string supplied or reported on PDB '
            'mismatches the ligand structure.'
        )
        warnings.warn(msg, LigandBondOrderWarning)
        return

def create_probe_mol(probe, use_conf=True):
    edmol = Chem.EditableMol(Chem.Mol())
    rd_atoms = {}
    conf = Chem.Conformer()
    conf.Set3D(True)
    for atom in probe.atoms:
        rd_atom = Chem.Atom(atom.element.capitalize())
        atom_idx = edmol.AddAtom(rd_atom)
        rd_atoms[atom.name] = atom_idx
        coord = Point3D(*atom.coord)
        conf.SetAtomPosition(atom_idx, coord)
    for bond in probe.bonds:
        a1, a2 = bond[0].name, bond[1].name
        idx1, idx2 = rd_atoms[a1], rd_atoms[a2]
        bo = get_rdkit_bond_order(bond.type)
        edmol.AddBond(idx1, idx2, bo)
    mol = edmol.GetMol()
    if use_conf:
        mol.AddConformer(conf)
    # AllChem.SanitizeMol(mol)
    # mol = AllChem.AddHs(mol, addCoords=use_conf)
    # AllChem.SanitizeMol(mol)
    # AllChem.ComputeGasteigerCharges(mol)
    # Chem.AssignStereochemistryFrom3D(mol)
    return mol

def create_probe_confomers(
        probe, conf_coords: np.ndarray, scores: np.ndarray = None
    ):
    if conf_coords.shape[1] != len(probe.atoms):
        raise ValueError(
            'Number of atoms in probe and number of coordinates do not match!'
        )
    
    if hasattr(probe, 'rdkit_mol'):
        # Make a copy of the stored rdkit mol, and remove all conf
        mol = Chem.RWMol(probe.rdkit_mol).GetMol()
        mol.RemoveAllConformers()
    else:
        mol = create_probe_mol(probe, use_conf=False)
    if scores is None:
        for conf_coord in conf_coords:
            conf = Chem.Conformer()
            conf.Set3D(True)
            for atom_idx, coord in enumerate(conf_coord):
                coord = Point3D(*coord.astype(np.float64))
                conf.SetAtomPosition(atom_idx, coord)
            mol.AddConformer(conf, assignId=True)
        return mol

    if scores.shape[0] != conf_coords.shape[0]:
        raise ValueError(
            'Number of conformers and number of scores do not match!'
        )
    for conf_coord, score in zip(conf_coords, scores):
        conf = Chem.Conformer()
        conf.Set3D(True)
        conf.SetDoubleProp('Score', float(score))
        for atom_idx, coord in enumerate(conf_coord):
            coord = Point3D(*coord.astype(np.float64))
            conf.SetAtomPosition(atom_idx, coord)
        mol.AddConformer(conf, assignId=True)
    return mol

def write_conformers_sdf(mol, filename):
    conformers = mol.GetConformers()
    if not conformers:
        raise ValueError('No conformers found in the molecule!')
    with AllChem.SDWriter(filename) as writer:
        for conf in conformers:
            writer.write(mol, confId=conf.GetId())


import json
import warnings
import requests
import numpy as np
from Bio.Align import PairwiseAligner
from crimm.Modeller.TopoLoader import ResidueTopologySet
from crimm.Superimpose.ChainSuperimposer import ChainSuperimposer
from crimm.Fetchers import fetch_rcsb, fetch_alphafold, uniprot_id_query
from crimm.StructEntities.Chain import PolymerChain

def find_gaps_within_range(gaps, segment):
    in_range = []
    for gap in gaps:
        if gap.issubset(segment):
            in_range.append(gap)
    return in_range

def find_gaps_range_overlap(gaps, segment):
    overlaps = []
    for gap in gaps:
        if segment.intersection(gap):
            overlaps.append(gap)
    return overlaps

def find_segment_offsets(chainA_seq, chainB_seq):
    aligner = PairwiseAligner()
    aligner.target_internal_open_gap_score = -100
    aligner.query_internal_open_gap_score = -100
    align = aligner.align(chainA_seq, chainB_seq)[0]
    chainA_aligned, chainB_aligned = align.aligned+1 # resseq is 1-indexed
    offsets = (chainB_aligned - chainA_aligned)[:,0]
    offset_dict = {
        range(*segmentA): offset for segmentA, offset in zip(
            chainA_aligned, offsets
        )
    }
    return offset_dict

def find_gap_offsets(gaps, segment_offset_dict):
    gap_offset_dict = {}
    for segment, offset in segment_offset_dict.items():
        segment = set(segment)
        gaps_in_range = find_gaps_within_range(gaps, segment)
        gap_offset_dict.update(
            {tuple(gap): offset for gap in gaps_in_range}
        )
        for gap in gaps_in_range:
            gaps.remove(gap)
        if len(gaps) == 0:
            break
    return gap_offset_dict

def translate_gap_ids(chainA, chainB):
    """Translate gap ids from chainA to chainB based on their canonical sequences 
    alignment"""
    if chainA.can_seq == chainB.can_seq:
        rp_gap = translate_gaps_with_identical_seqs(chainA, chainB)
        translated = {
            tuple(sorted(gap)):tuple(sorted(gap)) for gap in rp_gap
        }
        return translated

    segment_offsets = find_segment_offsets(chainA.can_seq, chainB.can_seq)
    gap_offsets = find_gap_offsets(chainA.gaps, segment_offsets)
    translated = {}
    for gap_A, offset in gap_offsets.items():
        translated_gap_A = {resseq+offset for resseq in gap_A}
        if find_gaps_range_overlap(chainB.gaps, translated_gap_A):
            continue
        translated[tuple(sorted(gap_A))] = tuple(sorted(translated_gap_A))
    return translated

def translate_gaps_with_identical_seqs(chainA, chainB):
    model_chain_gaps = chainA.gaps
    template_chain_gaps = chainB.gaps
    translated = []
    i, j = 0, 0

    while i<len(model_chain_gaps) and j<len(template_chain_gaps):
        # gap of the model chain
        m_gap = model_chain_gaps[i]
        # start idx of the gap
        m_start = min(m_gap)
        # gap of the template chain
        t_gap = template_chain_gaps[j]
        # end idx of the gap
        t_end =  max(t_gap)
             
################# Possible Scenarios for t_gap and m_gap ###################
#   ---------------[cur t_gap]------[next t_gap]-------  t_gap positions
# 1 ---------[cur m_gap]-------------------------------  overlap
# 2 --------------------[cur m_gap]--------------------  overlap
# 3 ----------------------------[cur m_gap]------------  m_start > p_end 
# 4 ---[cur m_gap]---------[next m_gap]----------------  repair cur m_gap
############################################################################

        # If gaps overlap, move on to the next gap for both (e.g 1 and 2)
        if m_gap.intersection(t_gap):
            i += 1
            j += 1
        # If the p_gap ends before the current model_gap,
        # move on to the next t_gap (e.g 3)
        elif m_start > t_end:
            j += 1
        # Otherwise, current model_gap can be fixed by the template_chain,
        # and move on to the next m_gap (e.g. 4)
        else:
            translated.append(m_gap)
            i += 1

    # If there are left-over gaps from receiving chain after going over 
    # the last gap of providing chain, these gaps can all be repaired.
    if i < len(model_chain_gaps):
        translated.extend(model_chain_gaps[i:])
    return translated

def report_loops(chain: PolymerChain):
    res_dict = dict(chain.missing_res)
    loops = []
    for gap in chain.gaps:
        res_ids = sorted(gap)
        missing_res = [res_dict[res_id] for res_id in res_ids]
        anchor, target = chain[min(gap)-1], chain[max(gap)+1]
        loops.append(
            {
                'anchor': {'C': anchor['C'].coord, 'CA': anchor['CA'].coord},
                'loop_res_ids': res_ids,
                'loop': missing_res,
                'target': {'N': target['N'].coord, 'CA': target['CA'].coord}
            }
        )
    return loops
class ChainLoopBuilder:
    """
    loop modeller for PDB protein structures by homology modeling
    """

    def __init__(
            self, model_chain: PolymerChain, 
            model_can_seq = None,
            pdbid = None,
            inplace = False,
        ):
        """
        Args:
            model_chain: the chain to be repaired
            model_can_seq (optional): the canonical sequence of the chain.
                If not provided, it will be assumed to be the canonical sequence
                parsed from mmcif file.
            pdbid (optional): the PDB ID of the chain. If not provided, it will
                look up the PDB ID from the chain's top parent 
                (model or structure containing the chain).
            inplace: if True, the model_chain will be modified in place. Default
                is False.
        """
        if not inplace:
            self.model_chain = model_chain.copy()
        else:
            self.model_chain = model_chain
        if model_can_seq is not None:
            self.model_can_seq = model_can_seq
        else:
            self.model_can_seq = model_chain.can_seq
        self.repaired_gaps = []
        self.repaired_residues = {}
        self.imposer =  ChainSuperimposer()
        self.template_chain = None
        self.template_can_seq = None
        self.translated_ids = None
        self.query_results = None
        self.pdbid = pdbid
        if self.pdbid is None:
            if (parent:=(model_chain.get_top_parent())).level == 'S':
                self.pdbid = parent.id
            elif (parent:=model_chain.get_top_parent()).level == 'M' and parent.pdb_id is not None:
                self.pdbid = parent.pdb_id
            else:
                warnings.warn(
                    f'PDB ID not set for {model_chain}! '
                    'Make sure the chain is loaded from a MMCIF file or from fetch() '
                    'to have the PDB ID correctly parsed, or set the PDB ID manually.'
                )

        if self.model_can_seq is None:
            raise AttributeError('Canonical sequence is required to repair loops')

    def set_template_chain(
            self, template_chain: PolymerChain
        ):
        """
        Set the template chain for copying the coordinates of loop residues from
        """
        self.template_chain = template_chain
        self.template_can_seq = template_chain.can_seq
        if self.template_can_seq is None:
            raise AttributeError(
                f'Canonical sequence does not exist for {template_chain}! '
                'Make sure the chain is loaded from a MMCIF file or from fetch() '
                'to have the canonical sequence correctly parsed.'
            )

        self.translated_ids = translate_gap_ids(
            self.model_chain, self.template_chain
        )

    def superimpose_two_chains(self, on_atoms = 'backbone'):
        """
        Impose template chain onto modeling chain
        Return the RMSD value
        """

        if not hasattr(self, 'template_chain'):
            raise AttributeError(
                'Template chain not set! Use set_template_chain() to load template first'
            )

        self.imposer.set_chains(
            self.model_chain, self.template_chain, on_atoms=on_atoms
        )
        self.imposer.apply_transform(self.template_chain)
        return self.imposer.rms

    def _copy_gap_residues(self, gap, residues):
        # gap is a set and has to be sorted before calculated the idx for 
        # insertion. Otherwise, it would mess up the residue sequence in
        # the structure

        residues = sorted(residues, key=lambda x: x.id[1])
        gap = sorted(gap)

        for i, res in zip(gap, residues):
            hetflag = res.id[0]
            res.id = (hetflag, i, ' ')
            # i in insert refer to child list index, thus use i-1 here
            self.model_chain.add(res)
        self.model_chain.sort_residues()
        self.repaired_gaps.append(tuple(gap))

    def get_repair_residues_from_template(
            self,
            rmsd_threshold = None,
        ):
        """
        Find repairable residue for model_chain from template_chain gaps.
        """
        if not hasattr(self, 'template_chain'):
            raise AttributeError(
                'Template chain not set!' 
                'Use set_template_chain() to load template first'
            )

        rmsd_qualified = {}
        for gap, template_resseqs in self.translated_ids.items():
            self.imposer.set_around_gap(
                self.model_chain, self.template_chain, gap, cutoff=5
            )
            self.imposer.apply_transform(self.template_chain)
            rmsd = self.imposer.rms
            if rmsd_threshold is not None and rmsd > rmsd_threshold:
                warnings.warn(
                    f'RMSD ({rmsd}) of superposition around gap {gap} is '
                    f'higher than the threshold {rmsd_threshold}. '
                    'Repair not applied!'
                )
                continue

            available_residues = []
            for i in template_resseqs:
                resseq = int(i)
                if resseq not in self.template_chain:
                    break
                available_residues.append(self.template_chain[resseq].copy())

            if len(available_residues) != len(template_resseqs):
                continue

            rmsd_qualified[gap] = {
                "residues":available_residues,
                'rmsd':rmsd
            }

        return rmsd_qualified

    def highlight_repaired_gaps(self, color = 'green', add_licorice = False):
        """
        Highlight the repaired gaps with color (default: cyan)
        """
        try:
            from crimm.Visualization.NGLVisualization import View
        except ImportError as exc:
            raise ImportError(
                "WARNING: nglview not found! Install nglview to show"
                "protein structures."
                "http://nglviewer.org/nglview/latest/index.html#installation"
            ) from exc
        fixed_res = []
        for res_list in self.repaired_gaps:
            for res_seq in res_list:
                fixed_res.append(self.model_chain[res_seq])
        if len(fixed_res) < 5:
            add_licorice = True # cartoon will not render if there are too few residues
        view = View()
        view.load_entity(self.model_chain)
        view.highlight_residues(
            fixed_res, color=color, add_licorice=add_licorice
        )

        return view

    def show_gaps(self):
        """
        Highlight the residues on the gap edges with color (default: red)
        """
        if len(self.model_chain.gaps) == 0:
            warnings.warn(f"No gap found in {self.model_chain}!")
        try:
            from crimm.Visualization.NGLVisualization import View
        except ImportError as exc:
            raise ImportError(
                "WARNING: nglview not found! Install nglview to show"
                "protein structures."
                "http://nglviewer.org/nglview/latest/index.html#installation"
            ) from exc
        view = View()

        view.load_entity(self.model_chain)
        view.subdue_all_entities()
        rgb_red = [1,0,0]
        line_width = 0.5
        for gap in self.model_chain.gaps:
            # Center on the gap by selecting the terminal residues around the gap
            if (
                (st:=min(gap)-1) in self.model_chain and 
                (end:=max(gap)+1) in self.model_chain
            ):
                st_coord = self.model_chain[st]['CA'].coord
                end_coord = self.model_chain[end]['CA'].coord
                arrow_name = f'gap:{st+1}-{end-1}'
                view.shape.add_arrow(
                    st_coord, end_coord, rgb_red, line_width, arrow_name
                )

        return view

    def show(self):
        from crimm.Visualization.NGLVisualization import show_nglview_multiple
        return show_nglview_multiple([self.model_chain, self.template_chain])

    @staticmethod
    def _build_seq_query(
            seq_string, max_num_match, identity_score_cutoff
        ):
        """
        Args:
            seq_string: str
            max_num_match: int
            identity_score_cutoff: float (from 0 to 1)
        """
        query_dict = {
            
        "query": {
            "type": "terminal",
            "service": "sequence",
            "parameters": {
            "evalue_cutoff": 1,
            "identity_cutoff": identity_score_cutoff,
            "sequence_type": "protein",
            "value": seq_string
            }
            },
        "request_options": {
            "scoring_strategy": "sequence",
            "paginate": {
                "start": 0,
                "rows": max_num_match
            }
        },
        "return_type": "polymer_instance"
        }
        
        return query_dict

    def query_seq_match(self, max_num_match, identity_score_cutoff):
        '''
        Make RCSB PDB query for protein sequence
        
        :param seq_string : string of the query sequence
        :param max_num_match : maximum number of top matches returned. 
        :param identity_score_cutoff: lowest identity score accepted
        :rtype retults : dict consists (sequence_matching_score, PDB_ID, entity_ID)
        '''
        seq_string = str(self.model_can_seq)
        json_q = ChainLoopBuilder._build_seq_query(
            seq_string, max_num_match, identity_score_cutoff
        )
        # entry point of rcsb.org for advancced search
        url = "https://search.rcsb.org/rcsbsearch/v2/query?"
        response = requests.post(url, json=json_q, timeout=500)
        r_dict = json.loads(response.text)

        if 'result_set' not in r_dict:
            raise ValueError(
                'Query on the sequence did not return '
                f'the requested information. {response}'
            )

        results = {}
        for entry in r_dict['result_set']:
            pdb_id, chain_id = entry['identifier'].split('.')
            if pdb_id == self.pdbid:
                continue
            if entry['score'] not in results:
                results[entry['score']] = {}
            if pdb_id not in results[entry['score']]:
                results[entry['score']][pdb_id] = []
            results[entry['score']][pdb_id].append(chain_id)
        return results

    @staticmethod
    def get_templates(query_results: dict, local_entry_point: str):
        """
        Return a generator of all template chains from the query result dict
        """
        for entity_dict in query_results.values():
            for pdbid, chain_ids in entity_dict.items():
                first_model = fetch_rcsb(
                    pdbid, use_bio_assembly = False,
                    include_solvent = False, 
                    local_entry=local_entry_point,
                    first_model_only=True
                )
                for chain_id in chain_ids:
                    template_chain = first_model[chain_id]
                    yield (pdbid, template_chain)

    def _reject_by_rmsd(self, overall_rmsd_threshold):
        if overall_rmsd_threshold is not None:
            rms = self.superimpose_two_chains()
            if rms > overall_rmsd_threshold:
                return True
        return False
            
    def _update_candidates(self, cur_repairables, candidates, pdbid):
        for gap_key, info_dict in cur_repairables.items():
            if gap_key not in candidates or (
                candidates[gap_key]['rmsd'] > info_dict['rmsd']
            ):  
                info_dict['structure_id'] = pdbid
                candidates[gap_key] = info_dict

    def build_from_homology(
            self,
            max_num_match = 10,
            identity_score_cutoff = 0.95,
            rmsd_threshold = None,
            overall_rmsd_threshold = None,
            local_entry_point = None
        ):
        """Build the missing loops of the model chain from homology models from
        PDB. Seqeunce search query is performed first to find the closest homology
        models based on sequence identity scores."""
        # if local_entry_point is None, rcsb.org will be used
        if self.model_chain.is_continuous():
            warnings.warn(
                f"{self.model_chain} does not have any missing loops"
            )
            return

        self.query_results = self.query_seq_match(
            max_num_match, identity_score_cutoff
        )

        if not self.query_results:
            warnings.warn(
                f"No homology model found for {self.pdbid}-{self.model_chain.id} "
                "from sequence search!"
            )
            return

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            # create a generator for all the templates
            all_templates = self.get_templates(
                self.query_results, local_entry_point
            )

        repair_candidates = {}
        for pdbid, template in all_templates:
            self.set_template_chain(template)
            if self._reject_by_rmsd(overall_rmsd_threshold):
                continue
            if cur_repairables := self.get_repair_residues_from_template(
                    rmsd_threshold = rmsd_threshold
                ):
                self._update_candidates(cur_repairables, repair_candidates, pdbid)

        for gap in sorted(repair_candidates):
            copy_residues = repair_candidates[gap]['residues']
            self._copy_gap_residues(
                gap = gap,
                residues = copy_residues,
            )
        self.model_chain.reset_atom_serial_numbers()
        self.repaired_residues.update(repair_candidates)
    
    def build_from_alphafold(self, include_terminal = False):
        """Build the missing loops of the model chain from AlphaFold models. 
        The AlphaFold models are downloaded from the AlphaFold database. If 
        include_terminal is True, the missing terminal residues will be copied
        from the AlphaFold model."""
        
        if len(self.model_chain.missing_res) == 0:
            warnings.warn(
                f"{self.model_chain} does not have any missing residues"
            )
            return

        if self.pdbid is None:
            warnings.warn(
                "Missing looper not built due to no PDB ID provided "
                f"for {self.model_chain}."
            )
            return
        uniprot_id = uniprot_id_query(self.pdbid, self.model_chain.entity_id)
        af_struct = fetch_alphafold(uniprot_id)
        if af_struct is None:
            warnings.warn(
                "Missing looper not built due to no AlphaFold structure avaible "
                f"for {self.model_chain}."
            )
            return
        
        af_chain = af_struct.child_list[0].child_list[0]
        self.set_template_chain(af_chain)
        repairables = self.get_repair_residues_from_template(
            rmsd_threshold = None
        )

        if len(repairables) == 0:
            warnings.warn(
                f"No repairable residues found for {self.model_chain} "
                f"from AlphaFold model {af_struct.id}."
            )
            return

        if not include_terminal:
            missing_terminals = []
            for gap in repairables:
                if 1 in gap or len(self.model_chain.can_seq) in gap:
                    missing_terminals.append(gap)
            # remove alphafold terminal residues from repairables
            for gap in missing_terminals:
                repairables.pop(gap)
            self.model_can_seq = self.model_chain.can_seq

        for gap_ids, repairable_info_dict in repairables.items():
            repairable_info_dict['structure_id'] = af_struct.id
            copy_residues = repairable_info_dict['residues']
            self._copy_gap_residues(
                gap = gap_ids,
                residues = copy_residues,
            )
        self.model_chain.reset_atom_serial_numbers()
        self.repaired_residues.update(repairables)

    def get_chain(self, return_incomplete_chain = False):
        """If no missing loop exists or return_incomplete_chain is True, this 
        function returns the chain that is being modelled. Otherwise, 
        it raises an error."""
        if not self.model_chain.is_continuous() and not return_incomplete_chain:
            raise ValueError(
                'The chain being worked on remains broken! If you have tried to repair the missing loops, '
                'this means that the method(s) failed on at least one missing loops.'
            )

        return self.model_chain
    
class ArcLoopBuilder:
    """
    Simple loop modeller that builds missing loops as arcs avoiding clashes
    with existing protein atoms.
    """
    def __init__(self, loop, protein_atoms):
        """
        start_residue: dict {'C': np.array, 'CA': np.array} of the residue BEFORE loop
        end_residue:   dict {'N': np.array, 'CA': np.array} of the residue AFTER loop
        sequence:      string of amino acids in the missing loop
        protein_atoms: list of np.array coordinates (the 'walls' to avoid)
        """
        self.topo = ResidueTopologySet('protein')
        self.start = loop['anchor']
        self.end = loop['target']
        self.loop = loop['loop']
        self.protein_atoms = protein_atoms
        self.num_residues = len(loop['loop'])
        self.res_ids = loop['loop_res_ids']
        # Standard geometries for approximations
        self.CA_CA_DIST = 3.8  # Avg distance between Alpha Carbons
        self.CLASH_DIST = 1.8  # Min distance to existing atoms

    def get_bezier_point(self, t, p0, p1, p2):
        """
        Quadratic Bezier formula: B(t) = (1-t)^2 * P0 + 2(1-t)t * P1 + t^2 * P2
        """
        return (1-t)**2 * p0 + 2*(1-t)*t * p1 + t**2 * p2

    def generate_ca_trace(self, apex_point):
        """
        Generates C-alpha coordinates along the Bezier curve defined by 
        Start, Apex, and End.
        """
        p0 = self.start['CA']
        p2 = self.end['CA']
        p1 = apex_point
        
        ca_coords = []
        # We need N+2 points (Start + N loop residues + End) to interpolate correctly
        # We only return the N loop residues.
        steps = self.num_residues + 1
        
        for i in range(1, steps):
            t = i / steps
            coord = self.get_bezier_point(t, p0, p1, p2)
            ca_coords.append(coord)
            
        return ca_coords

    def reconstruct_backbone(self, ca_trace):
        """
        Simple heuristic to place N and C atoms relative to the C-alphas.
        Since this is an *initial* conformation, we place them on the vectors
        connecting the C-alphas.
        """
        full_loop = []
        
        # Include anchors for vector calculation
        all_cas = [self.start['CA']] + ca_trace + [self.end['CA']]
        
        for i in range(self.num_residues):
            # Indices in all_cas: Current residue is i+1
            prev_ca = all_cas[i]
            curr_ca = all_cas[i+1]
            next_ca = all_cas[i+2]
            
            # Place N approx 1/3 way from Prev_CA to Curr_CA
            vec_n = (curr_ca - prev_ca)
            n_coord = prev_ca + vec_n * 0.76  # 0.76 is roughly (dist(CA-N) / 3.8)
            
            # Place C approx 1/3 way from Curr_CA to Next_CA
            vec_c = (next_ca - curr_ca)
            c_coord = curr_ca + vec_c * 0.39  # 0.39 is roughly (dist(CA-C) / 3.8)

            coord_dict = {
                'N': np.array(n_coord),
                'CA': np.array(curr_ca),
                'C': np.array(c_coord)
            }
            resname = self.loop[i]
            residue = topo[resname].create_residue_from_coord_dict(
                coord_dict, resseq = self.res_ids[i]
            )
            full_loop.append(residue)
            
        return full_loop
    
    def check_clashes(self, backbone):
        """Returns True if any atom in the new backbone hits the protein."""
        for res in backbone:
            for atom_type in ['N', 'CA', 'C']:
                coord = res[atom_type].coord
                for p_atom in self.protein_atoms:
                    if np.linalg.norm(coord - p_atom) < self.CLASH_DIST:
                        return True
        return False

    def build(self):
        # 1. Define the Axis (Chord)
        start_pt = self.start['CA']
        end_pt = self.end['CA']
        chord_vec = end_pt - start_pt
        chord_len = np.linalg.norm(chord_vec)
        midpoint = start_pt + (chord_vec / 2.0)
        
        # 2. Determine Loop Height
        # The loop must be longer than the chord. 
        # Approx length of loop = Num_Residues * 3.8
        target_len = self.num_residues * self.CA_CA_DIST
        
        if target_len < chord_len:
            print("Warning: Sequence too short to bridge gap. Stretching straight.")
            h = 0.1
        else:
            # Approx height calculation (Pythagoras on the triangle approximation)
            # This is a heuristic estimation for Bezier control point height
            h = np.sqrt((target_len/2.0)**2 - (chord_len/2.0)**2) * 1.5
            
        # 3. Define Perpendicular Vectors for Rotation
        # We need an arbitrary vector perpendicular to the chord to start rotating
        # Try Z-axis, if chord is Z-axis, try X-axis
        arbitrary = np.array([0, 0, 1])
        if np.allclose(chord_vec / chord_len, arbitrary):
            arbitrary = np.array([1, 0, 0])
            
        # Gram-Schmidt process to get a true perpendicular vector
        perp_vec = np.cross(chord_vec, arbitrary)
        perp_vec = perp_vec / np.linalg.norm(perp_vec)
        
        # 4. Swing the Loop (Rotate 0 to 360 degrees)
        best_backbone = None
        min_clashes = float('inf')
        
        # We test 36 angles (every 10 degrees)
        for angle in np.linspace(0, 2*np.pi, 36):
            
            # Rotate the perpendicular vector around the chord vector
            # (Rodrigues' rotation formula or simple trig reconstruction)
            
            # Simplified rotation logic:
            # Create a local coordinate system at the midpoint
            # u = chord (normalized), v = perp, w = cross(u, v)
            u = chord_vec / chord_len
            v = perp_vec
            w = np.cross(u, v)
            
            # Rotated direction
            rot_dir = v * np.cos(angle) + w * np.sin(angle)
            
            # The "Apex" (Control Point) for the Bezier curve
            apex = midpoint + (rot_dir * h)
            
            # Generate the trace
            ca_trace = self.generate_ca_trace(apex)
            
            # Build full atoms
            backbone = self.reconstruct_backbone(ca_trace)
            
            # Check collisions
            if not self.check_clashes(backbone):
                print(f"Found clear path at angle {int(np.degrees(angle))} degrees!")
                return backbone
            
        print("Could not find clash-free path. Returning best guess (0 degrees).")
        # In a real scenario, you might return the path with minimum clashes
        # Re-generate base case
        apex = midpoint + (perp_vec * h)
        return self.reconstruct_backbone(self.generate_ca_trace(apex))


"""Model class, used in Structure objects."""
import warnings
from Bio.PDB.Model import Model as _Model
from crimm.Utils.StructureUtils import index_to_letters

class Model(_Model):
    """The extended Model class representing a model in a structure.
    Derived from Biopython's Bio.PDB.Model and compatible with Biopython functions

    In a structure derived from an X-ray crystallography experiment,
    only a single model will be present (with some exceptions). NMR
    structures normally contain many different models.
    """
    def __init__(self, id, serial_num=None):
        super().__init__(id, serial_num=serial_num)
        self.pdb_id = None
        self.pdbx_description = None
        self.connect_dict = {}
        self.connect_atoms = {}

    def set_pdb_id(self, pdb_id):
        """Set the PDB ID of this model."""
        if self.pdb_id is not None:
            warnings.warn(
                f"Overwriting PDB ID {self.pdb_id} with {pdb_id}"
            )
        self.pdb_id = pdb_id

    def __repr__(self):
        return f"<Model id={self.get_id()} Chains={len(self)}>" 
    
    def expanded_view(self):
        """Print the hierarchy tree of this model."""
        hierarchy_str = repr(self)
        branch_symbols = '\n\t│\n\t├───'
        for chain in self:
            hierarchy_str += branch_symbols
            hierarchy_str += "\n\t├──────".join(chain.expanded_view().split('\n  '))
        return hierarchy_str

    def _ipython_display_(self):
        """Return the nglview interactive visualization window"""
        if len(self) == 0:
            return
        from crimm.Visualization import show_nglview_multiple
        from IPython.display import display
        display(show_nglview_multiple(self.child_list))
        print(self.expanded_view())

    @property
    def chains(self):
        """Alias for child_list. Returns the list of chains in this model."""
        return self.child_list
    
    def get_top_parent(self):
        if self.parent is None:
            return self
        return self.parent
    
    def reset_atom_serial_numbers(self, include_alt=True):
        """Reset all atom serial numbers in the encompassing entity (the parent 
        structure, if it exists) starting from 1."""
        i = 1
        for atom in self.get_atoms(include_alt=include_alt):
            atom.set_serial_number(i)
            i+=1
    
    def get_atoms(self, include_alt=False):
        """Return a generator of all atoms from this model. If include_alt is True, the 
        disordered residues will be expanded and altloc of disordered atoms will be included."""
        for chain in self:
            yield from chain.get_atoms(include_alt=include_alt)

    def set_connect(self, connect_dict):
        """Set the connect_dict attribute of this model. connect_dict is a dictionary of
        pairs of atom identified chain_id, resname, resseq, atom_id and altloc for 
        disulfide bonds, covalent bonds, metal coordination, saltbridge, etc."""
        self.connect_dict = connect_dict
        for connect_type, records in self.connect_dict.items():
            self.connect_atoms[connect_type] = []
            for record in records:
                atom_pairs = self._process_connect_record(record)
                if len(atom_pairs) != 2:
                    warnings.warn(
                        f"Insufficient atoms for connect record {connect_type}"
                    )
                    continue
                self.connect_atoms[connect_type].append(tuple(atom_pairs))

    def _process_connect_record(self, record):
        atom_pairs = []
        for atom_info in record:
            chain_id = atom_info['chain']
            resseq = atom_info['resseq']
            resname = str(atom_info['resname'])
            altloc = atom_info['altloc']
            atom_id = atom_info['atom_id']
            if chain_id not in self:
                warnings.warn(
                    f"Chain {chain_id} not found in model {self.get_id()}"
                )
                continue
            chain = self[chain_id]
            if chain.chain_type == 'Heterogens':
                for het_res in chain:
                    if het_res.resname == resname:
                        residue = het_res
            elif chain.chain_type == 'Solvent':
                # We simply cannot find which atom the record is referring to
                warnings.warn(
                    f"Chain {chain_id} is a solvent chain, connect atom "
                    "assignment skipped!"
                )
                continue
            elif resseq not in chain:
                warnings.warn(
                    f"Residue {resseq} not found in chain {chain_id}"
                )
                continue
            else:
                residue = chain[resseq]
            if residue.resname != resname:
                warnings.warn(
                    f"Residue {resseq} in chain {chain_id} has name "
                    f"{residue.resname}, not {resname}"
                )
                continue
            atom = residue[atom_id]
            if altloc:
                if atom.is_disordered() and altloc in atom.child_dict:
                # If altloc is specified and the atom is disordered, select the correct altloc
                # Note: some altloc might not exist for a single model. 
                # E.g. 2GRZ's altloc B reflects a completely different model (model 1 is A and model 2 is B)
                    atom = atom.child_dict[altloc]
                else:
                    warnings.warn(
                        f"Altloc {altloc} not found for atom {atom_id} in "
                        f"residue {resseq} of chain {chain_id} from model {self.get_id()}"
                    )
                    continue
            atom_pairs.append(atom)
        return tuple(atom_pairs)
    
    def replace_chain(self, in_chain, out_chain_id):
        """Swap the chain in this model with the given chain."""
        if out_chain_id not in self:
            raise ValueError(f"Model {self} does not have chain {out_chain_id}")
        if in_chain in self.child_list:
            raise ValueError(f"Chain {in_chain} already in model {self}")
        if in_chain.id != out_chain_id:
            warnings.warn(
                f"Renaming chain ID: {in_chain.id} to {out_chain_id} for"
                f"{in_chain}"
            )
            in_chain.id = out_chain_id
        self.detach_child(out_chain_id)
        self.add(in_chain)
        
    def remove_chains(self, chain_ids):
        """Remove the chains with the given IDs."""
        removed_chains = []
        for chain_id in chain_ids:
            if chain_id in self:
                removed_chains.append(self[chain_id])
                self.detach_child(chain_id)
        return removed_chains

    def keep_chains(self, chain_ids):
        """Keep only the chains with the given IDs."""
        remove_chain_ids = []
        for chain in self.child_list:
            if chain.id not in chain_ids:
                remove_chain_ids.append(chain.id)
        return self.remove_chains(remove_chain_ids)
    
    def sort_chains(self, reset_id=False):
        """Sort the chains in this model by chain ID."""
        self.child_list.sort(key=lambda x: x.id)
        self.child_dict = {chain.id: chain for chain in self.child_list}
        if reset_id:
            for i, chain in enumerate(self.child_list):
                chain.id = index_to_letters(i)


"""
Module containing the parser class for constructing structures from coord CRD 
files from CHARMM output.

Read a list of atoms from a CHARMM CARD coordinate file (CRD_)
to build a basic biopython/crimm structure.  Reads atom ids (ATOMNO), 
atom names (TYPES), resids (RESID), residue numbers (RESNO), 
residue names (RESNames), segment ids (SEGID) and tempfactor (Weighting).  
Atom element and mass are determined by a lookup table derived from CHARMM36 
and CGENFF residue topology files (rtf).

Residues are detected through a change in resid or resnum, 
while segments are detected according to changes in segid. chains are based on 
segid. The chain ids are assigned based on the alphabet.

"""
import warnings
from string import ascii_uppercase
from collections import namedtuple
import numpy as np
from Bio.PDB.PDBExceptions import PDBConstructionWarning
from crimm.IO.StructureBuilder import StructureBuilder
from crimm.IO.PDBParser import protein_letters_3to1, nucleic_letters_3to1
from crimm.Data.element_dict import all_element_dict
from crimm.IO.PDBParser import convert_chains

crd_entry = namedtuple(
    'crd_entry',
    [
        'serial', 'resnum', 'resname', 'atomname', 
        'coord', 'segid', 'resid', 'tempFactor'
    ]
)

class CRDParser:
    """Parse a CHARMM CARD coordinate file for structure coord information.

    Reads the following Attributes:
     - Atomids
     - Atomnames
     - Tempfactors
     - Resids
     - Resnames
     - Resnums
     - Segids

    Determines the following Attributes:
     - Atomtypes
     - Masses
    """
    def __init__(self, include_solvent = True, QUIET = False) -> None:
        self._structure_builder = StructureBuilder()
        self.QUIET = QUIET
        self.include_solvent = include_solvent

    @staticmethod
    def determine_chain_id(i, cur_letters=''):
        cur_letters = ascii_uppercase[i % 26] + cur_letters
        if (j := i//26) > 0:
            return CRDParser.determine_chain_id(j-1, cur_letters)
        return cur_letters

    def get_structure(self, filepath, structure_id = None):
        """Return the structure.

        Arguments:
         :structure_id: string, the id that will be used for the structure
         :filepath: path to mmCIF file, OR an open text mode file handle

        """
        with warnings.catch_warnings():
            if self.QUIET:
                warnings.filterwarnings(
                    "ignore", category=PDBConstructionWarning
                )
            entries = self.create_namedtuples(filepath)
            if structure_id is None:
                structure_id = filepath.split('/')[-1].split('.')[0]
            self._build_structure(structure_id, entries)

        return self._structure_builder.get_structure()

    def create_namedtuples(self, filepath):
        """Create a list of namedtuples from the CRD file."""
        with open(filepath, 'r') as f:
            lines = [l.rstrip() for l in f.readlines() if not l.startswith('*')]
        entries = []
        for l in lines[1:]:
            entry = l.split()
            (
                serial, resnum, resname, atomname, 
                x, y, z, segid, resid, tempFactor
            ) = entry
            coord = np.array([float(x), float(y), float(z)])
            if resname == 'ILE' and atomname == 'CD':
                atomname = 'CD1'
            cur_entry = crd_entry(
                int(serial), int(resnum), resname, atomname, 
                coord, segid, int(resid), float(tempFactor)
            )
            entries.append(cur_entry)
        return entries
    
    def _build_structure(self, structure_id, entries):
        sb = self._structure_builder
        sb.init_structure(structure_id)
        # Only one model per structure for crd files
        sb.init_model(1)
        # temporarily set to empty dict
        # TODO: Implement a header parser
        sb.header = {} 
        cur_segid = None
        cur_resid = None
        cur_resnum = None
        chain_id_dict = {}
        for entry in entries:
            if entry.segid != cur_segid:
                cur_segid = entry.segid
                sb.init_seg(cur_segid)
                chain_id = self.determine_chain_id(len(chain_id_dict))
                chain_id_dict[chain_id] = cur_segid
                sb.init_chain(chain_id)
            if entry.resid != cur_resid or entry.resnum != cur_resnum:
                cur_resid = entry.resid
                cur_resnum = entry.resnum
                if entry.resname not in protein_letters_3to1 and entry.resname not in nucleic_letters_3to1:
                    field = 'H'
                elif entry.resname in ('HOH', 'WAT', 'TIP3', 'TIP4'):
                    field = 'W'
                else:
                    field = ' '
                sb.init_residue(entry.resname, field, cur_resid, ' ')
            atom_name = entry.atomname
            if atom_name in all_element_dict:
                element = all_element_dict.get(atom_name)
            else:
                element = atom_name[0]
                warnings.warn(
                    f'Element type cannot be determined from atom name {atom_name}! '
                    f'Element is assigned as \'{element}\'.'
                )
                
            sb.init_atom(
                atom_name, entry.coord, entry.tempFactor,
                occupancy = 1.0,
                altloc = ' ',
                fullname = atom_name,
                serial_number = entry.serial,
                element = element
            )
        stucture = sb.structure
        for model in stucture:
            new_chains = convert_chains(model.child_list)
            if not self.include_solvent:
                new_chains = [c for c in new_chains if c.chain_type != 'Solvent']
            for chain in new_chains:
                chain.set_parent(model)
            model.child_list = new_chains
            model.child_dict = {c.id: c for c in new_chains}

        


from crimm.Fetchers import fetch_rcsb
from crimm.Modeller import ParameterLoader, TopologyLoader
from crimm.Modeller.LoopBuilder import ChainLoopBuilder
from crimm.Modeller.TopoFixer import fix_chain
from crimm.StructEntities.Model import Model
import crimm.Adaptors.pyCHARMMAdaptors as pcm_interface
from crimm.Adaptors.PropKaAdaptors import PropKaProtonator

from pycharmm.psf import delete_atoms as pcm_del_atoms
from pycharmm.psf import get_natom as pcm_get_natom
from pycharmm.generate import patch as pcm_patch
from pycharmm.settings import set_verbosity as pcm_set_verbosity

def minimize_chain(chain, sd_nstep, abnr_nstep):
    """Minimize crimm protein chain in CHARMM"""
    # load into CHARMM to minimize the structure
    if pcm_get_natom() > 0:
        pcm_del_atoms()
    pcm_interface.load_chain(chain)
    pcm_interface.minimize(sd_nstep=sd_nstep, abnr_nstep=abnr_nstep)
    # Uodate the coordinate in crimm structure
    pcm_interface.sync_coords(chain)

def correct_prot_first_patch(chain, default):
    """Correct the first patch name for PRO and GLY"""
    # PRO and GLY need special treatment when patched at the N-terminus 
    first_resname = chain.residues[0].resname
    if first_resname == 'PRO':
        first_patch = 'PROP'
    elif first_resname == 'GLY':
        first_patch = 'GLYP'
    else:
        first_patch = default
    return first_patch

def get_param_loaders():
    """get crimm rtf and prm loader for protein and nucleic acid chains
    also load them into CHARMM"""
    rtf_loader = {
        'prot': TopologyLoader('protein'),
        'na': TopologyLoader('nucleic')
    }
    param_loader = {
        'prot': ParameterLoader('protein'),
        'na': ParameterLoader('nucleic')
    }

    # fill the missing ic table values in the respective rtf
    for i, (chain_type, cur_rtf) in enumerate(rtf_loader.items()):
        cur_param = param_loader[chain_type]
        cur_param.fill_ic(cur_rtf)
        # load the respective files into CHARMM as well
        prev_level = pcm_set_verbosity(0)
        pcm_interface.load_topology(cur_rtf, append=bool(i))
        pcm_interface.load_parameters(cur_param, append=bool(i))
        pcm_set_verbosity(prev_level)
    return rtf_loader, param_loader

def protonate_and_patch(model, pH, rtf_loader, param_loader):
    """protonate all protein chains in a model and also load in charmm and
    patch the residues"""
    protonator = PropKaProtonator(
        rtf_loader, param_loader, pH = pH
    )
    protonator.load_model(model)
    protonator.apply_patches()

    if pcm_get_natom() > 0:
        pcm_del_atoms()
    for chain in model:
        if chain.id in protonator.patches and len(protonator.patches[chain.id]) > 0:
            built_atoms = fix_chain(chain)
        pcm_interface.load_chain(chain)

    for chain_id, patch_dict in protonator.patches.items():
        for resid, patch_name in patch_dict.items():
            pcm_patch(patch_name, f'PRO{chain_id} {resid}')

def load_pdbid_in_charmm(
    pdb_id,
    pH = 7.4,
    prot_first_patch = 'ACE',
    prot_last_patch = 'CT3',
    na_first_patch = '5TER',
    na_last_patch = '3PHO',
    sd_nstep = 300,
    abnr_nstep = 0,
    charmm_verbosity_level = 0,
):

    model = fetch_rcsb(
        pdb_id,
        include_solvent=False,
        # any existing hydrogen will be removed and rebuilt later
        include_hydrogens=False,
        first_model_only=True
    )

    prot_chains = {}
    na_chains = {} 
    # get the first model's id
    model_id = model.id
    # create a new empty model to store chains of interests
    new_model = Model(model_id)
    rtf_loader, param_loader = get_param_loaders()
    
    for chain in structure.models[0].chains:
        if chain.chain_type == 'Polypeptide(L)':
            prot_chains[chain.id] = chain
        elif chain.chain_type  == 'Polyribonucleotide':
            na_chains[chain.id] = chain

    for chain_id, chain in prot_chains.items():
        need_minimization = False
        # Missing loop in the chain
        if not chain.is_continuous():
            loop_builder = ChainLoopBuilder(chain)
            # Coordinates of the missing residues will be copied from
            # Alphafold structures
            # only build the loop not the termini
            loop_builder.build_from_alphafold(include_terminal = False)
            chain = loop_builder.get_chain()
            prot_chains[chain_id] = chain
            need_minimization = True
        prot_first_patch = correct_prot_first_patch(chain, default = prot_first_patch)
        rtf_loader['prot'].generate_chain_topology(
            chain,
            first_patch=prot_first_patch, 
            last_patch=prot_last_patch,
            # Coerce any modified residue to canonical residue that it is based on
            coerce=True
        )
        param_loader['prot'].fill_ic(rtf_loader['prot'])
        param_loader['prot'].apply(chain.topo_elements)
        fix_chain(chain)
        if need_minimization:
            # load into CHARMM to minimize the structure
            prev_level = pcm_set_verbosity(charmm_verbosity_level)
            minimize_chain(chain, sd_nstep, abnr_nstep)
            pcm_set_verbosity(prev_level)
        new_model.add(chain)

    for chain_id, chain in na_chains.items():
        # Missing loop is very unlikely in nucleotide chains on PDB
        # but if it exsits, an error will be raise
        if not chain.is_continuous():
            raise ValueError(
                f'Nucleotide chain {chain.id} is not continuous, '
                'topology cannot be generated.'
            )
        rtf_loader['na'].generate_chain_topology(
            chain, 
            first_patch=na_first_patch,
            last_patch=na_last_patch,
            coerce=True
        )
        param_loader['na'].fill_ic(rtf_loader['na'])
        param_loader['na'].apply(chain.topo_elements)
        fix_chain(chain)
        new_model.add(chain)
    # copy the connection record to the new model
    new_model.set_connect(structure.models[0].connect_dict)
    # replace the model with the new model in structure
    structure.detach_child(model_id)
    structure.add(new_model)
    protonate_and_patch(
        new_model, pH, rtf_loader['prot'], param_loader['prot']
    )
    # patch disulfide bonds if exists
    if 'disulf' in structure.models[0].connect_dict:
        for res1, res2 in structure.models[0].connect_dict['disulf']:
            seg1, seg2 = res1['chain'], res2['chain']
            seq1, seq2 = res1['resseq'], res2['resseq']
            patch_arg = f'PRO{seg1} {seq1} PRO{seg2} {seq2}'
            pcm_patch('DISU', patch_arg)

    return structure

if __name__ == '__main__':
    import warnings
    import argparse
    from pycharmm import write
    parser = argparse.ArgumentParser(
        prog="CHARMM Loader",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=
'''
  CHARMM Structure Loader from PDB ID
    Load a structure fetched from rcsb with PDB ID. Any missing loop will be constructed
    from the coordinates of the Alphafold structure. Protonation is done with PropKa model.
    Only protein and RNA chain will be processed, and only protein chain will be protonated
    with the specified pH value.
'''
    )
    parser.add_argument(
        'pdbid', metavar='P', type=str, help='PDB ID for the structure'
    )
    parser.add_argument(
        '-o','--outpath', type=str, default="./",
        help='Directory path for pdb and psf file outputs. '
        'Default to current directory'
    )
    parser.add_argument(
        '-p', '--ph', type=float, default=7.4,
        help='pH value for the protonation state for protein'
        'Default to pH=7.4.'
    )
    parser.add_argument(
        '-m', '--mini', type=int, default=300,
        help='Minimization steps to take after building any missing loop. '
        'Default to 300 steps of steepest descent.'
    )
    parser.add_argument(
        '--report', action='store_true',
        help='If the report flag is present, the crimm structure information '
        'will also be displayed.'
    )
    args = parser.parse_args()
    warnings.filterwarnings("ignore")
    # test_ids = ['5IEV', '1A8I', '7ZAP']
    structure = load_pdbid_in_charmm(
        args.pdbid,
        pH = args.ph,
        prot_first_patch = 'ACE',
        prot_last_patch = 'CT3',
        na_first_patch = '5TER',
        na_last_patch = '3PHO',
        sd_nstep = args.mini,
        abnr_nstep = 0,
        charmm_verbosity_level = 0,
    )
    write.coor_pdb(f'{args.pdbid}.pdb')
    write.psf_card(f'{args.pdbid}.psf')
    
    if args.report:
        print(structure)
        total_atoms = len(list(structure.get_atoms()))
        total_segments = len(structure.models[0])
        total_residues = 0
        total_bonds = 0
        total_angles = 0
        total_dihe = 0
        total_impr = 0
        for chain in structure.models[0].chains:
            print(chain.topo_elements)
            total_residues += len(chain)
            total_bonds += len(chain.topo_elements.bonds)
            total_angles += len(chain.topo_elements.angles)
            total_dihe += len(chain.topo_elements.dihedrals)
            total_impr += len(chain.topo_elements.impropers)
        print(
            'Report from crimm (Total)',
            f'segments={total_segments}, residues={total_residues}, atoms={total_atoms}, '
            f'bonds={total_bonds}, angles={total_angles}, dihedrals={total_dihe}, '
            f'impropers={total_impr}'
        )
    


# Description: Check if CUDA is available on the system
# -*- coding: utf-8 -*-

"""
Outputs some information on CUDA-enabled devices on your computer,
including current memory usage.
It's a port of https://gist.github.com/f0k/0d6431e3faa60bffc788f8b4daa029b1
from C to Python with ctypes, so it can run without compiling anything. Note
that this is a direct translation with no attempt to make the code Pythonic.
It's meant as a general demonstration on how to obtain CUDA device information
from Python without resorting to nvidia-smi or a compiled Python extension.
Author: Jan Schlüter, Ziqiao "Truman" Xu (徐梓乔)
License: MIT (https://gist.github.com/f0k/63a664160d016a491b2cbea15913d549#gistcomment-3870498)
"""

import sys
import ctypes
import warnings

# Some constants taken from cuda.h
CUDA_SUCCESS = 0
CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT = 16
CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_MULTIPROCESSOR = 39
CU_DEVICE_ATTRIBUTE_CLOCK_RATE = 13
CU_DEVICE_ATTRIBUTE_MEMORY_CLOCK_RATE = 36
# Table to convert CUDA architectural version number into core count per SM,
# data from https://github.com/NVIDIA/cuda-samples/blob/master/Common/helper_cuda.h
CUDA_CORE_TABLE = {
    # (major, minor): cores
    (1, 0): 8,    # Tesla
    (1, 1): 8,
    (1, 2): 8,
    (1, 3): 8,
    (2, 0): 32,   # Fermi
    (2, 1): 48,
    (3, 0): 192,  # Kepler
    (3, 2): 192,
    (3, 5): 192,
    (3, 7): 192,
    (5, 0): 128,  # Maxwell
    (5, 2): 128,
    (5, 3): 128,
    (6, 0): 64,   # Pascal
    (6, 1): 128,
    (6, 2): 128,
    (7, 0): 64,   # Volta
    (7, 2): 64,   # Xavier
    (7, 5): 64,   # Turing
    (8, 0): 64,   # Ampere
    (8, 6): 128,
    (8, 7): 128,
    (8, 9): 128,   # Ada
    (9, 0): 128,   # Hopper
}

def convert_SMVer_cores(major, minor):
    # Returns the number of CUDA cores per multiprocessor for a given
    # Compute Capability version. There is no way to retrieve that via
    # the API, so it needs to be hard-coded.
    # See _ConvertSMVer2Cores in helper_cuda.h in NVIDIA's CUDA Samples.
    return CUDA_CORE_TABLE.get((major, minor), 0)

def is_cuda_available():
    libnames = ('libcuda.so', 'libcuda.dylib', 'cuda.dll')
    for libname in libnames:
        try:
            cuda = ctypes.CDLL(libname)
        except OSError:
            continue
        else:
            break
    else:
        return False

    result = ctypes.c_int()
    result = cuda.cuInit(0)
    if result == CUDA_SUCCESS:
        return True

    return False


class CUDADeviceProp:
    """Class to contain property information about a single CUDA device.
    Should not be instantiated directly, but rather through the CUDAInfo class
    
    Attributes
    ----------
    device_id : int
        The device ID of the device
    name : str
        The name of the device
    cc_major : int
        The major compute capability version of the device
    cc_minor : int
        The minor compute capability version of the device
    compute_capability : str
        The compute capability version of the device
    cores : int
        The number of cores on the device
    threads_per_core : int
        The number of threads per core on the device
    clockrate : float
        The clockrate of the device in MHz
    mem_clockrate : float
        The memory clockrate of the device in MHz
    free_mem : float
        The current amount of free memory on the device in MiB
    total_mem : float
        The total amount of memory on the device in MiB
    """

    # class to contain cudaDeviceProp struct
    # https://docs.nvidia.com/cuda/cuda-runtime-api/structcudaDeviceProp.html
    def __init__(self, device_id: int, cuda_api: ctypes.CDLL) -> None:
        self._device_id = ctypes.c_int()
        self._get_error_str = cuda_api.cuGetErrorString

        if result := cuda_api.cuDeviceGet(
            ctypes.byref(self._device_id), device_id
        ) != CUDA_SUCCESS:
            error_str = ctypes.c_char_p()
            self._get_error_str(result, ctypes.byref(error_str))
            raise RuntimeError(
                f"Device {self.device_id} cuDeviceGet failed with error code {result}: ",
                f"{error_str.value.decode()}"
            )

        name = b' ' * 100
        cc_major = ctypes.c_int()
        cc_minor = ctypes.c_int()
        cores = ctypes.c_int()
        threads_per_core = ctypes.c_int()
        clockrate = ctypes.c_int()
        mem_clockrate = ctypes.c_int()
        self._context = ctypes.c_void_p()
        freeMem = ctypes.c_size_t()
        totalMem = ctypes.c_size_t()

        if result := cuda_api.cuDeviceGetName(
            ctypes.c_char_p(name), len(name), self._device_id
        ) != CUDA_SUCCESS:
            self._reveal_error_code(result, 'cuDeviceGetName')
            self.name = None
        else:
            name = name.decode()
            self.name = name.rstrip().rstrip('\x00')

        if result := cuda_api.cuDeviceComputeCapability(
            ctypes.byref(cc_major), ctypes.byref(cc_minor), self._device_id
        ) != CUDA_SUCCESS:
            self._reveal_error_code(result, 'cuDeviceComputeCapability')
            self.cc_major, self.cc_minor = None, None
        else:
            self.cc_major, self.cc_minor = cc_major.value, cc_minor.value

        if result := cuda_api.cuDeviceGetAttribute(
            ctypes.byref(cores),
            CU_DEVICE_ATTRIBUTE_MULTIPROCESSOR_COUNT,
            self._device_id
        ) != CUDA_SUCCESS:
            self._reveal_error_code(result, 'cuDeviceGetAttribute')
            self.cores = None
            self.cuda_cores = None
        else:
            self.cores = cores.value
            self.cuda_cores = self.cores * convert_SMVer_cores(self.cc_major, self.cc_minor) or "unknown"
            if result := cuda_api.cuDeviceGetAttribute(
                ctypes.byref(threads_per_core),
                CU_DEVICE_ATTRIBUTE_MAX_THREADS_PER_MULTIPROCESSOR, 
                self._device_id
            ) != CUDA_SUCCESS:
                self._reveal_error_code(result, 'cuDeviceGetAttribute')
                self.threads_per_core = None
                self.concurrent_threads = None
            else:
                self.threads_per_core = threads_per_core.value
                self.concurrent_threads = self.cores * self.threads_per_core
                
        if result := cuda_api.cuDeviceGetAttribute(
            ctypes.byref(clockrate),
            CU_DEVICE_ATTRIBUTE_CLOCK_RATE,
            self._device_id
        ) != CUDA_SUCCESS:
            self._reveal_error_code(result, 'cuDeviceGetAttribute')
            self.clockrate = None
        else:
            self.clockrate = clockrate.value / 1000. #MHz
        
        if result := cuda_api.cuDeviceGetAttribute(
            ctypes.byref(mem_clockrate),
            CU_DEVICE_ATTRIBUTE_MEMORY_CLOCK_RATE, 
            self._device_id
        ) != CUDA_SUCCESS:
            self._reveal_error_code(result, 'cuDeviceGetAttribute')
            self.mem_clockrate = None
        else:
            self.mem_clockrate = mem_clockrate.value / 1000. #MHz
        
        if hasattr(cuda_api, 'cuCtxCreate_v2'):
            create_context = cuda_api.cuCtxCreate_v2
            self._get_mem_info = cuda_api.cuMemGetInfo_v2
        else:
            create_context = cuda_api.cuCtxCreate
            self._get_mem_info = cuda_api.cuMemGetInfo
        self._destroy_context = cuda_api.cuCtxDetach
        
        if result := create_context(
            ctypes.byref(self._context), 0, self._device_id
        ) != CUDA_SUCCESS:
            self._reveal_error_code(result, 'cuCtxCreate')
            self._free_mem = None
            self.total_mem = None
        elif result := self._get_mem_info(
            ctypes.byref(freeMem), ctypes.byref(totalMem)
        ) != CUDA_SUCCESS:
            self._reveal_error_code(result, 'cuMemGetInfo')
            self._free_mem = None
            self.total_mem = None
        else:
            self._free_mem = freeMem.value / 1024**2 #MB
            self.total_mem = totalMem.value / 1024**2 #MB
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        self.detach()

    def detach(self):
        """Free the memory for CUDA context and destroy handle"""
        if self._context is not None:
            self._destroy_context(self._context)
            self._context = None

    def _reveal_error_code(self, result: int, func_name: str) -> None:
        error_str = ctypes.c_char_p()
        self._get_error_str(result, ctypes.byref(error_str))
        warnings.warn(
            f"Device {self.device_id} {func_name} failed with error code {result}: "
            f"{error_str.value.decode()}"
        )

    def _update_mem(self):
        if self._context is None:
            self._free_mem = None
            return

        free_mem = ctypes.c_size_t()
        total_mem = ctypes.c_size_t()

        if result := self._get_mem_info(
            ctypes.byref(free_mem),
            ctypes.byref(total_mem)
        ) != CUDA_SUCCESS:
            self._reveal_error_code(result, 'cuMemGetInfo')
            self._free_mem = None
            return
        self._free_mem = free_mem.value / 1024**2 #MB

    @property
    def free_mem(self):
        """Returns the free memory on the device in MiB."""
        self._update_mem()
        return self._free_mem

    @property
    def device_id(self):
        return self._device_id.value

    @property
    def compute_capability(self):
        """Returns the compute capability of the device."""
        return '.'.join([str(self.cc_major), str(self.cc_minor)])

    @property
    def is_detached(self):
        return self._context is None

    def __str__(self) -> str:
        info_str = (
            f"<Device {self.device_id}: {self.name} "
            f"totalMem={self.total_mem} MiB"
        )
        if self.free_mem is not None:
            info_str += f", freeMem={self.free_mem} MiB"
        info_str += ">"
        return info_str
    
    def __repr__(self) -> str:
        return str(self)
    
    def report(self) -> None:
        """Prints a report of the device's information."""
        print(f"<Device {self.device_id}: {self.name}>")
        print(f"\tCompute Capability: {self.compute_capability}")
        print(f"\tCUDA Cores: {self.cuda_cores}")
        print(f"\tClockrate: {self.clockrate} MHz")
        print(f"\tMemory Clockrate: {self.mem_clockrate} MHz")
        print(f"\tConcurrent Threads: {self.concurrent_threads}")
        print(f"\tTotal Memory: {self.total_mem} MiB")
        if self.free_mem is not None:
            print(f"\tFree Memory: {self.free_mem} MiB")


class CUDAInfo:
    """Container for available CUDA device information.
    Note, the CUDA driver API must be installed for this to work. The class will 
    attempt to load the CUDA driver API library from the default locations for 
    the current platform. If the library cannot be found, OSError will be 
    raised. If the library is found, but the CUDA driver API cannot be loaded
    from it, RuntimeError will be raised. 
    While using this utility for free memory monitoring, the class will allocate
    about 220 MiB of memory on each device for memory monitoring. This memory 
    will be deallocated when the context is detached. Thus, it is recommended to 
    use a context manager; the memory will be automatically freed when the 
    context is exited.

    Attributes:
        devices (list): List of CUDA devices available on the system.
        n_gpus (int): Number of CUDA devices available on the system.

    Usage:
    >>> from crimm.Utils.cuda_info import CUDAInfo
    ## with context manager
    >>> with CUDAInfo() as cuda_info:
    >>>     cuda_info.report()
    <CUDA Device Info for 1 GPU(s)>
    <Device 0: NVIDIA GeForce GTX 1080 Ti>
        Compute Capability: 6.1
        CUDA Cores: 3584
        Clockrate: 1582.0 MHz
        Memory Clockrate: 5505.0 MHz
        Concurrent Threads: 57344
        Total Memory: 11172.1875 MiB
        Free Memory: 11032.125 MiB
    ## or without context manager
    >>> cuda_info = CUDAInfo()
    >>> cuda_info.devices
    [<CUDA Device 0: NVIDIA GeForce GTX 1080 Ti>]
    >>> cuda_info.n_gpus
    1
    >>> cuda_info.devices[0].name # or cuda_info[0].name
    'NVIDIA GeForce GTX 1080 Ti'
    >>> cuda_info.devices[0].compute_capability
    '6.1'
    >>> cuda_info.devices[0].free_mem #MiB of free memory
    11032.125
    >>> cuda_info.detach() #deallocate CUDA memory and destroy CUDA context on all devices
    >>> cuda_info.devices[0].free_mem
    None
    >>> cuda_info.devices[0].is_detached
    True
    """

    # possible CUDA driver API library names
    # for linux, it's usually located in /usr/lib64/libcuda.so 
    libnames = ('libcuda.so', 'libcuda.dylib', 'cuda.dll')
    def __init__(self) -> None:
        self.devices = []
        self.cuda_api = self._find_cuda_api_lib()
        init_result = ctypes.c_int()
        init_result = self.cuda_api.cuInit(0)
        if init_result != CUDA_SUCCESS:
            error_str = ctypes.c_char_p()
            self.cuda_api.cuGetErrorString(init_result, ctypes.byref(error_str))
            raise OSError(
                "cuInit failed with error code "
                f"{init_result}: {error_str.value.decode()}"
            )

        self.n_gpus = self._get_num_gpus()
        self._enumerate_devices()

    def _find_cuda_api_lib(self):
        for libname in self.libnames:
            try:
                return ctypes.CDLL(libname)
            except OSError:
                continue
            else:
                break
        else:
            raise OSError(
                "Could not find CUDA driver. "
                "Failed to load any of: " + ' '.join(self.libnames)
            )

    def _get_num_gpus(self):
        nGpus = ctypes.c_int()
        result = ctypes.c_int()
        result = self.cuda_api.cuDeviceGetCount(ctypes.byref(nGpus))
        if result == CUDA_SUCCESS:
            return nGpus.value
        error_str = ctypes.c_char_p()
        self.cuda_api.cuGetErrorString(result, ctypes.byref(error_str))
        raise RuntimeError(
            "cuDeviceGetCount failed with error code "
            f"{result}: {error_str.value.decode()}"
        )

    def _enumerate_devices(self):
        for i in range(self.n_gpus):
            device = CUDADeviceProp(i, self.cuda_api)
            self.devices.append(device)

    def __str__(self) -> str:
        return f"<CUDAInfo: {self.n_gpus} GPUs>"
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        for device in self.devices:
            device.__exit__(exc_type, exc_value, traceback)

    def __del__(self):
        for device in self.devices:
            device.detach()

    def __getitem__(self, index):
        return self.devices[index]
    
    def __iter__(self):
        return iter(self.devices)
    
    def __len__(self):
        return len(self.devices)
    
    def report(self):
        print(f"<CUDA Device Info for {self.n_gpus} GPU(s)>")
        for device in self.devices:
            device.report()
    
    def detach(self):
        for device in self.devices:
            device.detach()

def report_cuda_devices():
    """Prints a report of the CUDA devices on the system."""
    with CUDAInfo() as cuda_info:
        cuda_info.report()
    return 0

if __name__=="__main__":
    sys.exit(report_cuda_devices())